#!/usr/bin/env python
# code extracted from: http://rosettacode.org/wiki/S-Expressions
# Originally taken from: https://gitlab.com/kicad/libraries/kicad-library-utils/-/blob/master/common/sexpr.py

import re

dbg = False

term_regex = r'''(?mx)
    \s*(?:
        (?P<brackl>\()|
        (?P<brackr>\))|
        (?P<num>[+-]?\d+\.\d+(?=[\ \)])|\-?\d+(?=[\ \)]))|
        (?P<sq>"(?:[^"]|(?<=\\)")*"(?:(?=\))|(?=\s)))|
        (?P<s>[^(^)\s]+)
       )'''

def parse_sexp(sexp):
    stack = []
    out = []
    if dbg: print("%-6s %-14s %-44s %-s" % tuple("term value out stack".split()))
    for termtypes in re.finditer(term_regex, sexp):
        term, value = [(t,v) for t,v in termtypes.groupdict().items() if v][0]
        if dbg: print("%-7s %-14s %-44r %-r" % (term, value, out, stack))
        if   term == 'brackl':
            stack.append(out)
            out = []
        elif term == 'brackr':
            assert stack, "Trouble with nesting of brackets"
            tmpout, out = out, stack.pop(-1)
            out.append(tmpout)
        elif term == 'num':
            v = float(value)
            if v.is_integer(): v = int(v)
            out.append(v)
        elif term == 'sq':
            out.append(value[1:-1].replace(r'\"', '"'))
        elif term == 's':
            out.append(value)
        else:
            raise NotImplementedError("Error: %r" % (term, value))
    assert not stack, "Trouble with nesting of brackets"
    return out[0]


"""The dimensions are used to mark spots in the board file for their dimensions (units, metric,
imperial, etc)

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    14.06.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_dimension
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List

from kiutils.items.common import Position
from kiutils.items.gritems import GrText
from kiutils.utils.strings import dequote

@dataclass
class DimensionFormat():
    """The ``format`` token defines the text formatting of a dimension

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_dimension_format
    """

    prefix: Optional[str] = None
    """The optional ``prefix`` token defines the string to add to the beginning of the dimension text"""

    suffix: Optional[str] = None
    """The optional ``suffix`` token defines the string to add to the end of the dimension text"""

    units: int = 3
    """The ``units`` token defines the dimension units used to display the dimension text. Valid units
    are as follows:
    - 0: Inches
    - 1: Mils
    - 2: Millimeters
    - 3: Automatic"""

    unitsFormat: int = 1
    """The ``unitsFormat`` token defines how the unit's suffix is formatted. Valid units formats are
    as follows:
    - 0: No suffix
    - 1: Bare suffix
    - 2: Wrap suffix in parenthesis"""

    precision: int = 4
    """The ``precision`` token defines the number of significant digits to display"""

    overrideValue: Optional[str] = None
    """The optional ``overrideValue`` token defines the text to substitute for the actual physical
    dimension"""

    suppressZeroes: bool = False
    """The ``suppressZeroes`` token removes all trailing zeros from the dimension text"""

    @classmethod
    def from_sexpr(cls, exp: list) -> DimensionFormat:
        """Convert the given S-Expresstion into a DimensionFormat object

        Args:
            - exp (list): Part of parsed S-Expression ``(format ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not format

        Returns:
            - DimensionFormat: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'format':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp[1:]:
            if type(item) != type([]):
                if item == 'suppress_zeroes': object.suppressZeroes = True
                continue
            if item[0] == 'prefix': object.prefix = item[1]
            if item[0] == 'suffix': object.suffix = item[1]
            if item[0] == 'units': object.units = item[1]
            if item[0] == 'units_format': object.unitsFormat = item[1]
            if item[0] == 'precision': object.precision = item[1]
            if item[0] == 'override_value': object.overrideValue = item[1]
            pass
        return object

    def to_sexpr(self, indent: int = 4, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        prefix = f' (prefix "{dequote(self.prefix)}")' if self.prefix is not None else ''
        suffix = f' (suffix "{dequote(self.suffix)}")' if self.suffix is not None else ''
        overwrite_val = f' (override_value "{dequote(self.overrideValue)}")' if self.overrideValue is not None else ''
        suppress_zeroes = f' suppress_zeroes' if self.suppressZeroes else ''

        expression =  f'{indents}(format{prefix}{suffix} (units {self.units}) (units_format {self.unitsFormat}) (precision {self.precision}){overwrite_val}{suppress_zeroes}){endline}'
        return expression

@dataclass
class DimensionStyle():
    """The ``style`` token defines the style of a dimension

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_dimension_style
    """

    thickness: float = 0.0
    """The ``thickness`` token defines the line thickness of the dimension"""

    arrowLength: float = 0.0
    """The ``arrowLength`` token defines the length of the dimension arrows"""

    textPositionMode: int = 0
    """The ``textPositionMode`` token defines the position mode of the dimension text. Valid position
    modes are as follows:
    - 0: Text is outside the dimension line
    - 1: Text is in line with the dimension line
    - 2: Text has been manually placed by the user"""

    extensionHeight: Optional[float] = None
    """The optional ``extensionHeight`` token defines the length of the extension lines past the
    dimension crossbar"""

    textFrame: Optional[int] = None
    """The optional ``textFrame`` token defines the style of the frame around the dimension text. This
    only applies to leader dimensions. Valid text frames are as follows:
    - 0: No text frame
    - 1: Rectangle
    - 2: Circle
    - 3:Rounded rectangle"""

    extensionOffset: Optional[float] = None
    """The optional ``extensionOffset`` token defines the distance from feature points to extension
    line start"""

    keepTextAligned: bool = False
    """The ``keepTextAligned`` token indicates that the dimension text should be kept in line with the
    dimension crossbar. When false, the dimension text is shown horizontally regardless of the
    orientation of the dimension."""

    @classmethod
    def from_sexpr(cls, exp: list) -> DimensionStyle:
        """Convert the given S-Expresstion into a DimensionStyle object

        Args:
            - exp (list): Part of parsed S-Expression ``(style ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not style

        Returns:
            - DimensionStyle: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'style':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp[1:]:
            if type(item) != type([]):
                if item == 'keep_text_aligned': object.keepTextAligned = True
                continue
            if item[0] == 'thickness': object.thickness = item[1]
            if item[0] == 'arrow_length': object.arrowLength = item[1]
            if item[0] == 'text_position_mode': object.textPositionMode = item[1]
            if item[0] == 'extension_height': object.extensionHeight = item[1]
            if item[0] == 'text_frame': object.textFrame = item[1]
            if item[0] == 'extension_offset': object.extensionOffset = item[1]
        return object

    def to_sexpr(self, indent: int = 4, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        extension_height = f' (extension_height {self.extensionHeight})' if self.extensionHeight is not None else ''
        text_frame = f' (text_frame {self.textFrame})' if self.textFrame is not None else ''
        extension_offset = f' (extension_offset {self.extensionOffset})' if self.extensionOffset is not None else ''
        keep_aligned = f' keep_text_aligned' if self.keepTextAligned else ''

        expression =  f'{indents}(style (thickness {self.thickness}) (arrow_length {self.arrowLength}) (text_position_mode {self.textPositionMode}){extension_height}{text_frame}{extension_offset}{keep_aligned}){endline}'
        return expression

@dataclass
class Dimension():
    """The ``dimension`` token defines a dimension in the PCB

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_dimension
    """

    locked: bool = False
    """The optional ``locked`` token specifies if the dimension can be moved"""

    type: str = "aligned"
    """The ``type`` token defines the type of dimension. Valid dimension types are ``aligned``,
    ``leader``, ``center``, ``orthogonal`` (and ``radial`` in KiCad version 7)"""

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the polygon resides on"""

    tstamp: Optional[str] = None
    """The ``tstamp`` token defines the unique identifier for the footprint. This only applies
    to footprints defined in the board file format."""

    pts: List[Position] = field(default_factory=list)
    """The ``pts`` token define the list of xy coordinates of the dimension"""

    height: Optional[float] = None
    """The optional ``height`` token defines the height of aligned dimensions"""

    orientation: Optional[float] = None
    """The optional ``orientation`` token defines the rotation angle for orthogonal dimensions"""

    leaderLength: Optional[float] = None
    """The optional ``leaderLength`` token attribute defines the distance from the marked radius to
    the knee for radial dimensions."""

    grText: Optional[GrText] = None
    """The optional ``grText`` token define the dimension text formatting for all dimension types
    except center dimensions"""

    format: Optional[DimensionFormat] = None
    """The optional ``format`` token define the dimension formatting for all dimension types except
    center dimensions"""

    style: DimensionStyle = field(default_factory=lambda: DimensionStyle())
    """The ``style`` token defines the dimension style information"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Dimension:
        """Convert the given S-Expresstion into a Dimension object

        Args:
            - exp (list): Part of parsed S-Expression ``(dimension ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not dimension

        Returns:
            - Dimension: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'dimension':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp[1:]:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                continue
            if item[0] == 'type': object.type = item[1]
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'height': object.height = item[1]
            if item[0] == 'orientation': object.orientation = item[1]
            if item[0] == 'leader_length': object.leaderLength = item[1]
            if item[0] == 'gr_text': object.grText = GrText().from_sexpr(item)
            if item[0] == 'format': object.format = DimensionFormat().from_sexpr(item)
            if item[0] == 'style': object.style = DimensionStyle().from_sexpr(item)
            if item[0] == 'pts':
                for point in item[1:]:
                    object.pts.append(Position().from_sexpr(point))

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Raises:
            - Exception: When number of coordinate points of the dimension equals 0

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        points = ''
        for point in self.pts:
            points = f'{points} (xy {point.X} {point.Y})'
        if len(points) == 0:
            raise Exception("Number of points must not be zero")

        expression =   f'{indents}(dimension (type {self.type}) (layer "{self.layer}") (tstamp {self.tstamp})\n'
        expression +=  f'{indents}  (pts{points})\n'
        if self.height is not None:
            expression +=  f'{indents}  (height {self.height})\n'
        if self.orientation is not None:
            expression +=  f'{indents}  (orientation {self.orientation})\n'
        if self.leaderLength is not None:
            expression +=  f'{indents}  (leader_length {self.leaderLength})\n'
        if self.grText is not None:
            expression += self.grText.to_sexpr(indent+2)
        if self.format is not None:
            expression += self.format.to_sexpr(indent+2)
        expression += self.style.to_sexpr(indent+2)
        expression +=  f'{indents}){endline}'
        return expression


"""Class to manage KiCad boards

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    20.02.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List, Dict
from os import path

from kiutils.items.common import Group, Image, Net, PageSettings, TitleBlock
from kiutils.items.zones import Zone
from kiutils.items.brditems import *
from kiutils.items.gritems import *
from kiutils.items.dimensions import Dimension
from kiutils.utils.strings import dequote
from kiutils.utils import sexpr
from kiutils.footprint import Footprint
from kiutils.misc.config import KIUTILS_CREATE_NEW_VERSION_STR, KIUTILS_CREATE_NEW_GENERATOR_STR

@dataclass
class Board():
    """The ``board`` token defines a KiCad layout according to the board file format used in
    ``.kicad_pcb`` files.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/
    """
    version: str = ""
    """The ``version`` token defines the board version using the YYYYMMDD date format"""

    generator: str = ""
    """The ``generator`` token defines the program used to write the file"""

    general: GeneralSettings = field(default_factory=lambda: GeneralSettings())
    """The ``general`` token defines general information about the board"""

    paper: PageSettings = field(default_factory=lambda: PageSettings())
    """The ``paper`` token defines informations about the page itself"""

    titleBlock: Optional[TitleBlock] = None
    """The ``titleBlock`` token defines author, date, revision, company and comments of the board"""

    layers: List[LayerToken] = field(default_factory=list)
    """The ``layers`` token defines all of the layers used by the board"""

    setup: SetupData = field(default_factory=lambda: SetupData())
    """The ``setup`` token is used to store the current settings used by the board"""

    properties: Dict[str, str] = field(default_factory=dict)
    """The ``properties`` token holds a list of key-value properties of the board as a dictionary"""

    nets: List[Net] = field(default_factory=list)
    """The ``nets`` token defines a list of nets used in the layout"""

    footprints: List[Footprint] = field(default_factory=list)
    """The ``footprints`` token defines a list of footprints used in the layout"""

    # TODO: Type hinting for this list
    graphicItems: List = field(default_factory=list) # as in gritems.py
    """The ``graphicItems`` token defines a list of graphical items used in the layout. Possible
    tokens are found in ``kiutils.items.gritems``
    
    The ``Image`` token is supported since KiCad v7 and must be added into this list when used."""

    traceItems: List = field(default_factory=list)
    """The ``traceItems`` token defines a list of segments, arcs and vias used in the layout"""

    zones: List[Zone] = field(default_factory=list)
    """The ``zones`` token defines a list of zones used in the layout"""

    dimensions: List[Dimension] = field(default_factory=list)
    """The ``dimensions`` token defines a list of dimensions on the PCB"""

    targets: List[Target] = field(default_factory=list)
    """The ``targets`` token defines a list of target markers on the PCB"""

    groups: List[Group] = field(default_factory=list)
    """The ``groups`` token defines a list of groups used in the layout"""

    filePath: Optional[str] = None
    """The ``filePath`` token defines the path-like string to the board file. Automatically set when
    ``self.from_file()`` is used. Allows the use of ``self.to_file()`` without parameters."""

    @classmethod
    def from_sexpr(cls, exp: list) -> Board:
        """Convert the given S-Expresstion into a Board object

        Args:
            - exp (list): Part of parsed S-Expression ``(kicad_pcb ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not kicad_pcb

        Returns:
            - Board: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'kicad_pcb':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'version': object.version = item[1]
            if item[0] == 'generator': object.generator = item[1]
            if item[0] == 'general': object.general = GeneralSettings().from_sexpr(item)
            if item[0] == 'paper': object.paper = PageSettings().from_sexpr(item)
            if item[0] == 'title_block': object.titleBlock = TitleBlock().from_sexpr(item)
            if item[0] == 'layers':
                for layer in item[1:]:
                    object.layers.append(LayerToken().from_sexpr(layer))
            if item[0] == 'setup': object.setup = SetupData().from_sexpr(item)
            if item[0] == 'property': object.properties.update({item[1]: item[2]})
            if item[0] == 'net': object.nets.append(Net().from_sexpr(item))
            if item[0] == 'footprint': object.footprints.append(Footprint().from_sexpr(item))
            if item[0] == 'gr_text': object.graphicItems.append(GrText().from_sexpr(item))
            if item[0] == 'gr_text_box': object.graphicItems.append(GrTextBox().from_sexpr(item))
            if item[0] == 'gr_line': object.graphicItems.append(GrLine().from_sexpr(item))
            if item[0] == 'gr_rect': object.graphicItems.append(GrRect().from_sexpr(item))
            if item[0] == 'gr_circle': object.graphicItems.append(GrCircle().from_sexpr(item))
            if item[0] == 'gr_arc': object.graphicItems.append(GrArc().from_sexpr(item))
            if item[0] == 'gr_poly': object.graphicItems.append(GrPoly().from_sexpr(item))
            if item[0] == 'gr_curve': object.graphicItems.append(GrCurve().from_sexpr(item))
            if item[0] == 'image': object.graphicItems.append(Image().from_sexpr(item))
            if item[0] == 'dimension': object.dimensions.append(Dimension().from_sexpr(item))
            if item[0] == 'target': object.targets.append(Target().from_sexpr(item))
            if item[0] == 'segment': object.traceItems.append(Segment().from_sexpr(item))
            if item[0] == 'arc': object.traceItems.append(Arc().from_sexpr(item))
            if item[0] == 'via': object.traceItems.append(Via().from_sexpr(item))
            if item[0] == 'zone': object.zones.append(Zone().from_sexpr(item))
            if item[0] == 'group': object.groups.append(Group().from_sexpr(item))

        return object

    @classmethod
    def from_file(cls, filepath: str, encoding: Optional[str] = None) -> Board:
        """Load a board directly from a KiCad board file (`.kicad_pcb`) and sets the
        ``self.filePath`` attribute to the given file path.

        Args:
            - filepath (str): Path or path-like object that points to the file
            - encoding (str, optional): Encoding of the input file. Defaults to None (platform 
                                        dependent encoding).

        Raises:
            - Exception: If the given path is not a file

        Returns:
            - Footprint: Object of the Schematic class initialized with the given KiCad schematic
        """
        if not path.isfile(filepath):
            raise Exception("Given path is not a file!")

        with open(filepath, 'r', encoding=encoding) as infile:
            item = cls.from_sexpr(sexpr.parse_sexp(infile.read()))
            item.filePath = filepath
            return item

    @classmethod
    def create_new(cls) -> Board:
        """Creates a new empty board with its attributes set as KiCad would create it

        Returns:
            - Board: Empty board
        """
        board = cls(
            version = KIUTILS_CREATE_NEW_VERSION_STR,
            generator = KIUTILS_CREATE_NEW_GENERATOR_STR
        )

        # Add all standard layers to board
        board.layers.extend([
            LayerToken(ordinal=0, name='F.Cu', type='signal'), 
            LayerToken(ordinal=31, name='B.Cu', type='signal'), 
            LayerToken(ordinal=32, name='B.Adhes', type='user', userName="B.Adhesive"),
            LayerToken(ordinal=33, name='F.Adhes', type='user', userName="F.Adhesive"),
            LayerToken(ordinal=34, name='B.Paste', type='user'),
            LayerToken(ordinal=35, name='F.Paste', type='user'),
            LayerToken(ordinal=36, name='B.SilkS', type='user', userName="B.Silkscreen"),
            LayerToken(ordinal=37, name='F.SilkS', type='user', userName="F.Silkscreen"),
            LayerToken(ordinal=38, name='B.Mask', type='user'),
            LayerToken(ordinal=39, name='F.Mask', type='user'),
            LayerToken(ordinal=40, name='Dwgs.User', type='user', userName="User.Drawings"),
            LayerToken(ordinal=41, name='Cmts.User', type='user', userName="User.Comments"),
            LayerToken(ordinal=42, name='Eco1.User', type='user', userName="User.Eco1"),
            LayerToken(ordinal=43, name='Eco2.User', type='user', userName="User.Eco2"),
            LayerToken(ordinal=44, name='Edge.Cuts', type='user'),
            LayerToken(ordinal=45, name='Margin', type='user'),
            LayerToken(ordinal=46, name='B.CrtYd', type='user', userName="B.Courtyard"),
            LayerToken(ordinal=47, name='F.CrtYd', type='user', userName="F.Courtyard"),
            LayerToken(ordinal=48, name='B.Fab', type='user'),
            LayerToken(ordinal=49, name='F.Fab', type='user'),
            LayerToken(ordinal=50, name='User.1', type='user'),
            LayerToken(ordinal=51, name='User.2', type='user'),
            LayerToken(ordinal=52, name='User.3', type='user'),
            LayerToken(ordinal=53, name='User.4', type='user'),
            LayerToken(ordinal=54, name='User.5', type='user'),
            LayerToken(ordinal=55, name='User.6', type='user'),
            LayerToken(ordinal=56, name='User.7', type='user'),
            LayerToken(ordinal=57, name='User.8', type='user'),
            LayerToken(ordinal=58, name='User.9', type='user')
        ])

        # Append net0 to netlist
        board.nets.append(Net())

        return board

    def to_file(self, filepath = None, encoding: Optional[str] = None):
        """Save the object to a file in S-Expression format

        Args:
            - filepath (str, optional): Path-like string to the file. Defaults to None. If not set, 
                                        the attribute ``self.filePath`` will be used instead.
            - encoding (str, optional): Encoding of the output file. Defaults to None (platform 
                                        dependent encoding).

        Raises:
            - Exception: If no file path is given via the argument or via `self.filePath`
        """
        if filepath is None:
            if self.filePath is None:
                raise Exception("File path not set")
            filepath = self.filePath

        with open(filepath, 'w', encoding=encoding) as outfile:
            outfile.write(self.to_sexpr())

    def to_sexpr(self, indent=0, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        addNewLine = False

        expression =  f'{indents}(kicad_pcb (version {self.version}) (generator {self.generator})\n\n'
        expression += self.general.to_sexpr(indent+2) + '\n'
        expression += self.paper.to_sexpr(indent+2)
        if self.titleBlock is not None:
            expression += self.titleBlock.to_sexpr(indent+2) + '\n'
        expression += f'{indents}  (layers\n'
        for layer in self.layers:
            expression += layer.to_sexpr(indent+4)
        expression += f'{indents}  )\n\n'
        expression += self.setup.to_sexpr(indent+2) + '\n'
        # Properties, if any
        if len(self.properties) > 0:
            for key, value in self.properties.items():
                expression += f'  (property "{dequote(key)}" "{dequote(value)}")\n'
            expression += '\n'

        # Nets
        if len(self.nets) > 0:
            for net in self.nets:
                expression += net.to_sexpr(indent=indent+2, newline=True)
            expression += '\n'

        # Footprints
        for footprint in self.footprints:
            expression += footprint.to_sexpr(indent+2, layerInFirstLine=True) + '\n'

        # Lines, Texts, Arcs and other graphical items
        if len(self.graphicItems) > 0:
            addNewLine = True
            for item in self.graphicItems:
                if isinstance(item, GrPoly):
                    expression += item.to_sexpr(indent+2, pts_newline=True)
                else:
                    expression += item.to_sexpr(indent+2)

        # Dimensions
        if len(self.dimensions) > 0:
            addNewLine = True
            for dimension in self.dimensions:
                expression += dimension.to_sexpr(indent+2)

        # Target markers:
        if len(self.targets) > 0:
            addNewLine = True
            for target in self.targets:
                expression += target.to_sexpr(indent+2)

        if addNewLine:
            expression += '\n'

        # Segments, vias and arcs
        if len(self.traceItems) > 0:
            for item in self.traceItems:
                expression += item.to_sexpr(indent+2)
            expression += '\n'

        # Zones
        for zone in self.zones:
            expression += zone.to_sexpr(indent+2)

        # Groups
        for group in self.groups:
            expression += group.to_sexpr(indent+2)

        expression += f'{indents}){endline}'
        return expression



"""Unittests of board related classes

Authors:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0
"""

import unittest
from os import path
from kiutils.footprint import Attributes

from tests.testfunctions import to_file_and_compare, prepare_test, cleanup_after_test, TEST_BASE
from kiutils.board import Board

BOARD_BASE = path.join(TEST_BASE, 'board')

class Tests_Board(unittest.TestCase):
    """Test cases for Boards"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_boardTraceArcs(self):
        """Tests the parser's handling of traces with arcs"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(BOARD_BASE, 'test_boardTraceArcs')
        board = Board().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(board, self.testData))

    def test_boardStackup32LayerDielectricsVias(self):
        """Tests the parsing of a board with 32 layers, all different dielectric layers and all
        available via combinations"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(BOARD_BASE, 'test_boardStackup32LayerDielectricsVias')
        board = Board().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(board, self.testData))

    def test_boardWithAllPrimitives(self):
        """Tests the parsing of a board containting all primitives (traces, texts, forms, dimensions,
        markers, polygons, etc)"""
        self.testData.pathToTestFile = path.join(BOARD_BASE, 'test_boardWithAllPrimitives')
        board = Board().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(board, self.testData))

    def test_allFpManufacturingAttributes(self):
        """Tests the parsing of a board with footprints that feature all combinations of
        manufacturing attributes. Tests all possible combinations of the following:
        <ul>
            <li>Board Only: True / False</li>
            <li>Exclude from BOM: True / False</li>
            <li>excludeFromPosFiles: True / False</li>
            <li>Type: SMD, THT, Other</li>
        </ul>

        Furthermore tests if the Attributes() object of a footprint is correctly created even
        when the parsed footprint has no (attr ...) token in its S-Expression."""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(BOARD_BASE, 'test_allFpManufacturingAttributes')
        board = Board().from_file(self.testData.pathToTestFile)

        # Test parsing
        self.assertTrue(to_file_and_compare(board, self.testData))

        # Test that attributes object is created, even when an empty attribute list is present in
        # parsed footprint
        attr = Attributes(boardOnly=False, excludeFromBom=False, excludeFromPosFiles=False, type=None)
        self.assertEqual(attr, board.footprints[11].attributes,
            msg="Parsing of footprint without `attr` field does not yield expected Attributes() object")

    def test_createEmptyBoard(self):
        """Tests the behavior when creating an empty board"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(BOARD_BASE, 'test_createEmptyBoard')
        board = Board().create_new()
        self.assertTrue(to_file_and_compare(board, self.testData))

    def test_footprintPadNewLines(self):
        """Renames the libId token (setting and unsetting) of footprints on a board"""
        self.testData.pathToTestFile = path.join(BOARD_BASE, 'test_renameFootprintIdToken')
        board = Board().from_file(self.testData.pathToTestFile)
        board.footprints[0].libId = "I_was_renamed:BUS_PCIexpress_x1"
        board.footprints[1].libId = "I_was_added:BUS_PCIexpress_x1"
        self.assertTrue(to_file_and_compare(board, self.testData))

    def test_pcbPlotParams(self):
        """Tests the parsing of board plot params"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(BOARD_BASE, 'test_pcbPlotParams')
        board = Board().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(board, self.testData))

class Tests_Board_Since_V7(unittest.TestCase):
    """Test cases for Boards since KiCad 7"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_textBoxAllVariants(self):
        """Tests all variants of the ``text_box`` token for text boxes in PCBs"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(BOARD_BASE, 'since_v7', 'test_textBoxAllVariants')
        board = Board().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(board, self.testData))

    def test_imageWithLayerToken(self):
        """Tests the new ``layer`` token for images in PCBs"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(BOARD_BASE, 'since_v7', 'test_imageWithLayerToken')
        board = Board().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(board, self.testData))

    def test_pcbPlotParams(self):
        """Tests the parsing of board plot params since KiCad v7"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(BOARD_BASE, 'since_v7', 'test_pcbPlotParams')
        board = Board().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(board, self.testData))
        
    def test_textsWithRenderCaches(self):
        """Tests text elements with the ``render_cache`` token"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(BOARD_BASE, 'since_v7', 'test_textsWithRenderCaches')
        board = Board().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(board, self.testData))


"""The zone token defines a zone on the board or footprint. Zones serve two purposes in
   KiCad: filled copper zones and keep out areas.

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    11.02.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_graphics_items
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List

from kiutils.items.common import Position
from kiutils.utils.strings import dequote

@dataclass
class KeepoutSettings():
    """The ``keepout `` token attributes define which objects should be kept out of the
    zone. This section only applies to keep out zones.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_zone_keep_out_settings
    """

    tracks: str = "allowed"
    """The ``tracks`` token attribute defines whether or not tracks should be excluded
    from the keep out area. Valid attributes are ``allowed`` and ``not_allowed``."""

    vias: str = "allowed"
    """The ``vias`` token attribute defines whether or not vias should be excluded from
    the keep out area. Valid attributes are ``allowed`` and ``not_allowed``."""

    pads: str = "allowed"
    """The ``pads`` token attribute defines whether or not pads should be excluded from
    the keep out area. Valid attributes are ``allowed`` and ``not_allowed``."""

    copperpour: str = "not-allowed"
    """The ``copperpour`` token attribute defines whether or not copper pours should be
    excluded from the keep out area. Valid attributes are ``allowed`` and ``not_allowed``."""

    footprints: str = "not-allowed"
    """The ``footprints`` token attribute defines whether or not footprints should be
    excluded from the keep out area. Valid attributes are ``allowed`` and ``not_allowed``."""

    @classmethod
    def from_sexpr(cls, exp: list) -> KeepoutSettings:
        """Convert the given S-Expresstion into a KeepoutSettings object

        Args:
            - exp (list): Part of parsed S-Expression ``(keepout ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not keepout

        Returns:
            - KeepoutSettings: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'keepout':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                continue

            if item[0] == 'tracks': object.tracks = item[1]
            if item[0] == 'vias': object.vias = item[1]
            if item[0] == 'pads': object.pads = item[1]
            if item[0] == 'copperpour': object.copperpour = item[1]
            if item[0] == 'footprints': object.footprints = item[1]

        return object

    def to_sexpr(self, indent: int = 0, newline: bool = False) -> str:
        """Generate the S-Expression representing this object. When no coordinates are set
        in the curve, the resulting S-Expression will be left empty.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        # KiCad seems to add a whitespace to the pad token here
        return f'{indents}(keepout (tracks {self.tracks}) (vias {self.vias}) (pads {self.pads} ) (copperpour {self.copperpour}) (footprints {self.footprints})){endline}'

@dataclass
class FillSettings():
    """The ``fill`` token attributes define how the zone is to be filled.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_zone_fill_settings
    """

    yes: bool = False
    """The ``yes`` token specifies if the zone should be filled. If not specified, the zone is
    not filled and no additional attributes are required."""

    mode: Optional[str] = None
    """The optional ``mode`` token attribute defines how the zone is filled. The only valid fill
    mode is ``hatched``. When not defined, the fill mode is solid."""

    thermalGap: Optional[float] = None
    """The optional ``thermalGap`` token attribute defines the distance from the zone to all
    pad thermal relief connections to the zone."""

    thermalBridgeWidth: Optional[float] = None
    """The optional ``thermalBridgeWidth`` token attribute defines the spoke width for all
    pad thermal relief connection to the zone."""

    smoothingStyle: Optional[str] = None
    """The optional ``smoothingStyle`` token attributes define the style of corner smoothing. Valid
    smoothing styles are ``chamfer`` and ``fillet``"""

    smoothingRadius: Optional[float] = None
    """The optional ``smoothingRadius`` token attributes define the radius of corner smoothing"""

    islandRemovalMode: Optional[int] = None
    """The optional ``islandRemovalMode`` token attribute defines the island removal mode.
    Valid island removal modes are:
    - 0: Always remove islands.
    - 1: Never remove islands.
    - 2: Minimum area island to allow.
    """

    islandAreaMin: Optional[float] = None
    """The optional ``islandAreaMin`` token attribute defines the minimum allowable zone
      island. This only valid when the remove islands mode is set to 2."""

    hatchThickness: Optional[float] = None
    """The optional ``hatchThickness`` token attribute defines the thickness for hatched fills"""

    hatchGap: Optional[float] = None
    """The optional ``hatchGap`` token attribute defines the distance between lines for hatched
    fills"""

    hatchOrientation: Optional[float] = None
    """The optional ``hatchOrientation`` token attribute defines the line angle for hatched fills"""

    hatchSmoothingLevel: Optional[int] = None
    """The optional ``hatchSmoothingLevel`` token attribute defines how hatch outlines are
    smoothed. Valid hatch smoothing levels are:
    - 0: No smoothing
    - 1: Fillet
    - 2: Arc minimum
    - 3: Arc maximum
    """

    hatchSmoothingValue: Optional[float] = None
    """The optional ``hatchSmoothingValue`` token attribute defines the ratio between the hole
    and the chamfer/fillet size"""

    hatchBorderAlgorithm: Optional[int] = None
    """The optional ``hatchBorderAlgorithm`` token attribute defines the if the zone line
    thickness is used when performing a hatch fill. Valid values for the hatch border
    algorithm are:
    - 0: Use zone minimum thickness.
    - 1: Use hatch thickness.
    """

    hatchMinHoleArea: Optional[float] = None
    """The optional ``hatchMinHoleArea`` token attribute defines the minimum area a hatch file hole can be"""

    @classmethod
    def from_sexpr(cls, exp: list) -> FillSettings:
        """Convert the given S-Expresstion into a FillSettings object

        Args:
            - exp (list): Part of parsed S-Expression ``(fill ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not fill

        Returns:
            - FillSettings: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'fill':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'yes': object.yes = True
                else: continue

            if item[0] == 'mode': object.mode = item[1]
            if item[0] == 'thermal_gap': object.thermalGap = item[1]
            if item[0] == 'thermal_bridge_width': object.thermalBridgeWidth = item[1]
            if item[0] == 'smoothing': object.smoothingStyle = item[1]
            if item[0] == 'radius': object.smoothingRadius = item[1]
            if item[0] == 'island_removal_mode': object.islandRemovalMode = item[1]
            if item[0] == 'island_area_min': object.islandAreaMin = item[1]
            if item[0] == 'hatch_thickness': object.hatchThickness = item[1]
            if item[0] == 'hatch_gap': object.hatchGap = item[1]
            if item[0] == 'hatch_orientation': object.hatchOrientation = item[1]
            if item[0] == 'hatch_smoothing_level': object.hatchSmoothingLevel = item[1]
            if item[0] == 'hatch_smoothing_value': object.hatchSmoothingValue = item[1]
            if item[0] == 'hatch_border_algorithm': object.hatchBorderAlgorithm = item[1]
            if item[0] == 'hatch_min_hole_area': object.hatchMinHoleArea = item[1]

        return object

    def to_sexpr(self, indent: int = 0, newline: bool = False) -> str:
        """Generate the S-Expression representing this object. When no coordinates are set
        in the curve, the resulting S-Expression will be left empty.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        yes = ' yes' if self.yes else ''
        mode = f' (mode {self.mode})' if self.mode is not None else ''
        smoothing = f' (smoothing {self.smoothingStyle})' if self.smoothingStyle is not None else ''
        radius = f' (radius {self.smoothingRadius})' if self.smoothingRadius is not None else ''
        irm = f' (island_removal_mode {self.islandRemovalMode})' if self.islandRemovalMode is not None else ''
        iam = f' (island_area_min {self.islandAreaMin})' if self.islandAreaMin is not None else ''
        ht = f'\n{indents}  (hatch_thickness {self.hatchThickness})' if self.hatchThickness is not None else ''
        hg = f' (hatch_gap {self.hatchGap})' if self.hatchGap is not None else ''
        ho = f' (hatch_orientation {self.hatchOrientation})' if self.hatchOrientation is not None else ''
        hsl = f'\n{indents}  (hatch_smoothing_level {self.hatchSmoothingLevel})' if self.hatchSmoothingLevel is not None else ''
        hsv = f' (hatch_smoothing_value {self.hatchSmoothingValue})' if self.hatchSmoothingValue is not None else ''
        hba = f'\n{indents}  (hatch_border_algorithm {self.hatchBorderAlgorithm})' if self.hatchBorderAlgorithm is not None else ''
        hmha = f' (hatch_min_hole_area {self.hatchMinHoleArea})' if self.hatchMinHoleArea is not None else ''

        return f'{indents}(fill{yes}{mode} (thermal_gap {self.thermalGap}) (thermal_bridge_width {self.thermalBridgeWidth}){smoothing}{radius}{irm}{iam}{ht}{hg}{ho}{hsl}{hsv}{hba}{hmha}){endline}'

@dataclass
class ZonePolygon():
    """The ``polygon`` token defines a list of coordinates that define part of a zone"""

    coordinates: List[Position] = field(default_factory=list)
    """The ``coordinates`` defines the list of polygon X/Y coordinates used to define the zone polygon"""

    @classmethod
    def from_sexpr(cls, exp: list) -> ZonePolygon:
        """Convert the given S-Expresstion into a ZonePolygon object

        Args:
            - exp (list): Part of parsed S-Expression ``(polygon ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not polygon

        Returns:
            - ZonePolygon: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'polygon':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                continue
            if item[0] == 'pts':
                for position in item[1:]:
                    object.coordinates.append(Position().from_sexpr(position))

        return object

    def to_sexpr(self, indent: int = 4, newline: bool = True) -> str:
        """Generate the S-Expression representing this object. When no coordinates are set
        in the polygon, the resulting S-Expression will be left empty.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object. If the polygon has no coordinates, an empty 
                   expression is returned.
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        if len(self.coordinates) == 0:
            return f'{indents}{endline}'

        expression =  f'{indents}(polygon\n'
        expression += f'{indents}  (pts\n'
        for point in self.coordinates:
            expression += f'{indents}    (xy {point.X} {point.Y})\n'
        expression += f'{indents}  )\n'
        expression += f'{indents})\n'
        return expression

@dataclass
class FilledPolygon():
    """The ``filled_polygon`` token defines the polygons used to fill a zone

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_zone_fill_polygons
    """

    layer: str = "F.Cu"
    """The ``layer`` token attribute defines the canonical layer the zone fill resides on"""

    # TODO: What is the definiton of this token?
    island: bool = False
    """The ``island`` token's definition has to be defined .."""

    coordinates: List[Position] = field(default_factory=list)
    """The ``coordinates`` defines the list of polygon X/Y coordinates used to fill the zone"""

    @classmethod
    def from_sexpr(cls, exp: list) -> FilledPolygon:
        """Convert the given S-Expresstion into a FilledPolygon object

        Args:
            - exp (list): Part of parsed S-Expression ``(filled_polygon ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not filled_polygon

        Returns:
            - FilledPolygon: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'filled_polygon':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                continue

            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'island': object.island = True
            if item[0] == 'pts':
                for position in item[1:]:
                    object.coordinates.append(Position().from_sexpr(position))

        return object

    def to_sexpr(self, indent: int = 4, newline: bool = True) -> str:
        """Generate the S-Expression representing this object. When no coordinates are set
        in the filled polygon, the resulting S-Expression will be left empty.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object. If the filled polygon has no coordinates, an empty 
                   expression is returned.
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        if len(self.coordinates) == 0:
            return f'{indents}{endline}'

        expression =  f'{indents}(filled_polygon\n'
        expression += f'{indents}  (layer "{dequote(self.layer)}")\n'
        if self.island:
            expression += f'{indents}  (island)\n'
        expression += f'{indents}  (pts\n'
        for point in self.coordinates:
            expression += f'{indents}    (xy {point.X} {point.Y})\n'
        expression += f'{indents}  )\n'
        expression += f'{indents})\n'
        return expression

# TODO: This is KiCad 4 stuff, has to be tested yet ..
@dataclass
class FillSegments():
    """The ``fill_polygon`` token defines the segments used to fill the zone. This is only
       used when loading boards prior to version 4 which filled zones with segments.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_zone_fill_segments
    """

    layer: str = "F.Cu"
    """The ``layer`` token attribute defines the canonical layer the zone fill resides on"""

    coordinates: List[Position] = field(default_factory=list)
    """The ``coordinates`` defines the list of polygon X/Y coordinates used to fill the zone."""

    @classmethod
    def from_sexpr(cls, exp: list) -> FillSegments:
        """Convert the given S-Expresstion into a FillSegments object

        Args:
            - exp (list): Part of parsed S-Expression ``(fill_segments ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not fill_segments

        Returns:
            - FillSegments: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'fill_segments':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                continue

            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'pts':
                for position in item[1:]:
                    object.coordinates.append(Position().from_sexpr(position))

        return object

    def to_sexpr(self, indent: int = 4, newline: bool = True) -> str:
        """Generate the S-Expression representing this object. When no coordinates are set
        in the curve, the resulting S-Expression will be left empty.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object. If the fill segments has no coordinates, an empty 
              expression is returned.
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        if len(self.coordinates) == 0:
            return f'{indents}{endline}'

        expression =  f'{indents}(fill_segments\n'
        expression += f'{indents}  (layer "{dequote(self.layer)}")\n'
        expression += f'{indents}  (pts\n'
        for point in self.coordinates:
            expression += f'{indents}    (xy {point.X} {point.Y})\n'
        expression += f'{indents}  )\n'
        expression += f'{indents})\n'
        return expression

@dataclass
class Hatch():
    """Data wrapper for Zone class hatching attribute"""

    style: str = "none"
    """The ``style`` token defines the style of the hatching. Valid hatch styles are ``none``, ``edge``
    and ``full``"""

    pitch: float = 0.0
    """The ``pitch`` token defines the pitch of the hatch"""

@dataclass
class Zone():
    """The ``zone`` token defines a zone on the board or footprint. Zones serve two purposes
       in KiCad: filled copper zones and keep out areas.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_zone

    """
    locked: bool = False
    """The ``locked`` token defines if the zone may be edited or not (Missing in KiCad
    docu as of 11.02.2022)"""

    net: int = 0
    """The ``net`` token attribute defines by the net ordinal number which net in the nets
    section that the zone is part of"""

    netName: str = "unknown"
    """The ``net_name`` token attribute defines the name of the net if the zone is not a keep
    out area. The net name attribute will be an empty string if the zone is a keep out area."""

    layers: List[str] = field(default_factory=list)
    """The ``layers`` token define the canonical layer set the zone connects as a list of
    strings. When the zone only resides on one layer, the output of ``self.to_sexpr()`` will
    change into ``(layer "xyz")`` instead of ``(layers ..)`` automatically."""

    tstamp: Optional[str] = None       # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the zone object"""

    name: Optional[str] = None
    """The optional ``name`` token attribute defines the name of the zone if one has been assigned"""

    hatch: Hatch = field(default_factory=lambda: Hatch())
    """The ``hatch`` token attributes define the zone outline display hatch style and pitch"""

    priority: Optional[int] = None
    """The optional ``priority`` attribute defines the zone priority if it is not zero"""

    connectPads: Optional[str] = None  # This refers to CONNECTION_TYPE in the docu
    """The ``connectPads`` token attributes define the pad connection type and clearance. Valid
    pad connection types are ``thru_hole_only``, ``full`` and ``no``. If the pad connection type is not
    defined, thermal relief pad connections are used"""

    clearance: float = 0.254
    """The ``clearance`` token defines the thermal relief for pad connections. The usage of this
    token is depending on the value of ``connectPads``."""

    minThickness: float = 0.254
    """The ``minThickness`` token attributed defines the minimum fill width allowed in the zone"""

    filledAreasThickness: Optional[str] = None
    """The optional ``filledAreasThickness`` attribute no specifies if the zone like width is
    not used when determining the zone fill area. This is to maintain compatibility with older
    board files that included the line thickness when performing zone fills when it is not defined."""

    keepoutSettings: Optional[KeepoutSettings] = None
    """The optional ``keepoutSettings`` section defines the keep out items if the zone
    defines as a keep out area"""

    fillSettings: Optional[FillSettings] = None
    """The optional ``fillSettings`` section defines how the zone is to be filled"""

    polygons: List[ZonePolygon] = field(default_factory=list)
    """The ``polygon`` token defines a list of zone polygons that define the shape of the zone"""

    filledPolygons: List[FilledPolygon] = field(default_factory=list)
    """The ``filledPolygons`` token defines a list of filled polygons in the zone"""

    # TODO: This is KiCad 4 only stuff, needs to be tested yet ..
    fillSegments: Optional[FillSegments] = None
    """The optional ``fillSegments`` section defines a list of track segments used to fill
    the zone"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Zone:
        """Convert the given S-Expresstion into a Zone object

        Args:
            - exp (list): Part of parsed S-Expression ``(zone ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not zone

        Returns:
            - Zone: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'zone':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                else: continue

            if item[0] == 'net': object.net = item[1]
            if item[0] == 'net_name': object.netName = item[1]
            if item[0] == 'layers' or item[0] == 'layer':
                for layer in item[1:]:
                    object.layers.append(layer)
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'name': object.name = item[1]
            if item[0] == 'hatch':
                object.hatch = Hatch(style=item[1], pitch=item[2])
            if item[0] == 'priority': object.priority = item[1]
            if item[0] == 'connect_pads':
                if len(item) == 2:
                    object.clearance = item[1][1]
                else:
                    object.connectPads = item[1]
                    object.clearance = item[2][1]
            if item[0] == 'min_thickness': object.minThickness = item[1]
            if item[0] == 'filled_areas_thickness': object.filledAreasThickness = item[1]
            if item[0] == 'keepout': object.keepoutSettings = KeepoutSettings().from_sexpr(item)
            if item[0] == 'fill': object.fillSettings = FillSettings().from_sexpr(item)
            if item[0] == 'polygon': object.polygons.append(ZonePolygon().from_sexpr(item))
            if item[0] == 'filled_polygon': object.filledPolygons.append(FilledPolygon().from_sexpr(item))
            if item[0] == 'fill_segments': object.fillSegments = FillSegments().from_sexpr(item)

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Raises:
            - Exception: When the zone has no elements in its layer list

        Returns:
            - str: S-Expression of this object.
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        locked = f' locked' if self.locked else ''
        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        name = f' (name "{dequote(self.name)}")' if self.name is not None else ''
        contype = f' {self.connectPads}' if self.connectPads is not None else ''
        fat = f' (filled_areas_thickness {self.filledAreasThickness})' if self.filledAreasThickness is not None else ''
        layers, layer_token = '', ''
        for layer in self.layers:
            layers += f' "{dequote(layer)}"'

        if len(self.layers) == 1:
            layer_token = f' (layer{layers})'
        elif len(self.layers) > 1:
            layer_token = f' (layers{layers})'
        else:
            raise Exception("Zone: No layers set for this zone")

        expression =  f'{indents}(zone{locked} (net {self.net}) (net_name "{dequote(self.netName)}"){layer_token}{tstamp}{name} (hatch {self.hatch.style} {self.hatch.pitch})\n'
        if self.priority is not None:
            expression += f'{indents}  (priority {self.priority})\n'
        expression += f'{indents}  (connect_pads{contype} (clearance {self.clearance}))\n'
        expression += f'{indents}  (min_thickness {self.minThickness}){fat}\n'
        if self.keepoutSettings is not None:
            expression += f'{indents}  {self.keepoutSettings.to_sexpr()}\n'
        if self.fillSettings is not None:
            expression += self.fillSettings.to_sexpr(indent+2, True)

        for polygon in self.polygons:
            expression += polygon.to_sexpr(indent+2)

        for polygon in self.filledPolygons:
            expression += polygon.to_sexpr(indent+2)

        # TODO: This is KiKad 4 stuff...
        if self.fillSegments is not None:
            expression += self.fillSegments.to_sexpr()
        expression += f'{indents}){endline}'
        return expression



"""Unittests of worksheet related classes

Authors:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0
"""

import unittest
from os import path

from tests.testfunctions import to_file_and_compare, prepare_test, cleanup_after_test, TEST_BASE
from kiutils.wks import WorkSheet

WORKSHEET_BASE = path.join(TEST_BASE, 'worksheets')

class Tests_WorkSheets(unittest.TestCase):
    """Test cases for Worksheets"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_allWorkSheetItems(self):
        """Tests the parsing of all available worksheet items"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(WORKSHEET_BASE, 'test_allWorkSheetItems')
        wks = WorkSheet().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(wks, self.testData))

    def test_createNewWorksheet(self):
        """Tests the ``create_new()`` function to create a new empty worksheet"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(WORKSHEET_BASE, 'test_createNewWorksheet')
        wks = WorkSheet.create_new()
        self.assertTrue(to_file_and_compare(wks, self.testData))



"""Classes for custom design rules (.kicad_dru) and its contents

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    26.06.2022 - created

Documentation taken from:
    ??? Syntax help in Pcbnew
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List
from os import path

from kiutils.utils import sexpr
from kiutils.utils.strings import dequote

@dataclass
class Constraint():
    """The ``Constraint`` token defines a design rule's constraint"""

    type: str = "clearance"
    """The ``type`` token defines the type of constraint. Defaults to ``clearance``. Allowed types:
    - ``annular_width`` - Width of an annular ring
    - ``clearance`` - Clearance between two items
    - ``courtyard_clearance`` - Clearance between two courtyards
    - ``diff_pair_gap`` - Gap between differential pairs
    - ``diff_pair_uncoupled`` - ???
    - ``disallow`` - ??? Do not allow this rule
    - ``edge_clearance`` - Clearance between the item and board edges
    - ``length`` - Length of the item
    - ``hole_clearance`` - Clearance between the item and holes
    - ``hole_size`` - Size of the holes associated with this item
    - ``silk_clearance`` - Clearance to silk screen
    - ``skew`` - Difference in length between the items associated with this constraint
    - ``track_width`` - Width of the tracks associated with this constraint
    - ``via_count`` - Number of vias
    - ``via_diameter`` - Diameter of vias associated with this constraint
    """

    min: Optional[str] = None
    """The ``min`` token defines the minimum allowed in this constraint"""

    opt: Optional[str] = None
    """The ``opt`` token defines the optimum allowed in this constraint"""

    max: Optional[str] = None
    """The ``max`` token defines the maximum allowed in this constraint"""

    elements: List[str] = field(default_factory=list)
    """The ``items`` token defines a list of zero or more element types to include in this constraint.
    The following element types are available:
    - ``buried_via``
    - ``micro_via``
    - ``via``
    - ``graphic``
    - ``hole``
    - ``pad``
    - ``text``
    - ``track``
    - ``zone``
    """

    @classmethod
    def from_sexpr(cls, exp: list) -> Constraint:
        """Convert the given S-Expresstion into a Constraint object

        Args:
            - exp (list): Part of parsed S-Expression ``(constraint ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the list's first parameter is not the ``(constraint ..)`` token

        Returns:
            - Constraint: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'constraint':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.type = exp[1]
        for item in exp[2:]:
            if type(item) != type([]):
                object.elements.append(item)
            if item[0] == 'min': object.min = item[1]
            if item[0] == 'opt': object.opt = item[1]
            if item[0] == 'max': object.max = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        min = f' (min "{dequote(self.min)}")' if self.min is not None else ''
        opt = f' (opt "{dequote(self.opt)}")' if self.opt is not None else ''
        max = f' (max "{dequote(self.max)}")' if self.max is not None else ''

        elements = ' '+' '.join(self.elements) if (len(self.elements) > 0) else ''

        return f'{indents}(constraint {self.type}{min}{opt}{max}{elements}){endline}'

@dataclass
class Rule():
    """The ``Rule`` token defines a custom design rule"""

    name: str = ""
    """The ``name`` token defines the name of the custom design rule"""

    constraints: List[Constraint] = field(default_factory=list)
    """The ``constraints`` token defines a list of constraints for this custom design rule"""

    condition: str = ""
    """The ``condition`` token defines the conditions that apply for this rule. Check KiCad syntax
    reference for more information. Example rule:
    - `A.inDiffPair('*') && !AB.isCoupledDiffPair()`"""

    layer: Optional[str] = None
    """The optional ``layer`` token defines the canonical layer the rule applys to"""

    severity: Optional[str] = None
    """The optional ``severity`` token defines the severity of the design rule. Valid values are
    ``warning``, ``error``, ``exclusion`` or ``ignore``.
    
    Available since KiCad v7"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Rule:
        """Convert the given S-Expresstion into a Rule object

        Args:
            - exp (list): Part of parsed S-Expression ``(rule ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the list's first parameter is not the ``(rule ..)`` token

        Returns:
            - Rule: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'rule':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.name = exp[1]
        for item in exp[2:]:
            if item[0] == 'constraint': object.constraints.append(Constraint().from_sexpr(item))
            if item[0] == 'condition': object.condition = item[1]
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'severity': object.severity = item[1]
        return object

    def to_sexpr(self, indent: int = 0):
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent

        expression = f'{indents}(rule "{dequote(self.name)}"\n'
        if self.layer is not None:
            expression += f'{indents}  (layer "{dequote(self.layer)}")\n'
        for item in self.constraints:
            expression += f'{indents}{item.to_sexpr(indent+2)}'
        expression += f'{indents}  (condition "{dequote(self.condition)}")'
        if self.severity is not None:
            expression += f'\n{indents}  (severity "{dequote(self.severity)}")'
        expression += ')\n'
        return expression

@dataclass
class DesignRules():
    """The ``DesignRules`` token defines a set of custom design rules (`.kicad_dru` files)"""

    version: int = 1
    """The ``version`` token defines the version of the file for the KiCad parser. Defaults to 1."""

    rules: List[Rule] = field(default_factory=list)
    """The ``rules`` token defines a list of custom design rules"""

    filePath: Optional[str] = None
    """The ``filePath`` token defines the path-like string to the schematic file. Automatically set when
    ``self.from_file()`` is used. Allows the use of ``self.to_file()`` without parameters."""

    @classmethod
    def from_sexpr(cls, exp: list) -> DesignRules:
        """Convert the given S-Expresstion into a DesignRules object

        Args:
            - exp (list): Part of parsed S-Expression ``(version ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the list's first parameter is not the ``(version ..)`` token

        Returns:
            - DesignRules: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if not isinstance(exp[0], list):
            raise Exception("Expression does not have the correct type")

        if exp[0][0] != 'version':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'version': object.version = item[1]
            if item[0] == 'rule': object.rules.append(Rule().from_sexpr(item))
        return object

    @classmethod
    def from_file(cls, filepath: str, encoding: Optional[str] = None) -> DesignRules:
        """Load a custom design rules set directly from a KiCad design rules file (`.kicad_dru`) and
        sets the ``self.filePath`` attribute to the given file path.

        Args:
            - filepath (str): Path or path-like object that points to the file
            - encoding (str, optional): Encoding of the input file. Defaults to None (platform 
                                        dependent encoding).
        Raises:
            - Exception: If the given path is not a file

        Returns:
            - Footprint: Object of the DesignRules class initialized with the given KiCad file
        """
        if not path.isfile(filepath):
            raise Exception("Given path is not a file!")

        with open(filepath, 'r', encoding=encoding) as infile:
            # This dirty fix adds opening and closing brackets `(..)` to the read input to enable
            # the S-Expression parser to work for the DRU-format as well.
            data = f'({infile.read()})'
            item = cls.from_sexpr(sexpr.parse_sexp(data))
            item.filePath = filepath
            return item

    @classmethod
    def create_new(cls) -> DesignRules:
        """Creates a new empty design rules set as KiCad would create it

        Returns:
            - DesignRules: Empty design rules set
        """
        return cls(version=1)

    def to_file(self, filepath = None, encoding: Optional[str] = None):
        """Save the object to a file in S-Expression format

        Args:
            - filepath (str, optional): Path-like string to the file. Defaults to None. If not set,
                                        the attribute ``self.filePath`` will be used instead.
            - encoding (str, optional): Encoding of the output file. Defaults to None (platform 
                                        dependent encoding).

        Raises:
            - Exception: If no file path is given via the argument or via `self.filePath`
        """
        if filepath is None:
            if self.filePath is None:
                raise Exception("File path not set")
            filepath = self.filePath

        with open(filepath, 'w', encoding=encoding) as outfile:
            outfile.write(self.to_sexpr())

    def to_sexpr(self, indent=0, newline=False):
        """Generate the S-Expression representing this object

        Args:
            indent (int, optional): Number of whitespaces used to indent the output. Defaults to 0.
            newline (bool, optional): Adds a newline to the end of the output. Defaults to False.

        Returns:
            str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression = f'{indents}(version {self.version})\n'

        if len(self.rules):
            expression += f'{indents}\n'
            for rule in self.rules:
                expression += f'{indents}{rule.to_sexpr(indent=indent)}'

        return expression + endline


"""Symbol graphical items define all of the drawing items that are used in the symbol
definition. This includes text, text boxes, lines, rectangles, circles, arcs, polygons
and curves.

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    16.02.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbol_graphic_items
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Optional

from kiutils.items.common import Fill, Position, Stroke, Effects, Fill
from kiutils.utils.strings import dequote

@dataclass
class SyArc():
    """The ``arc`` token defines a graphical arc in a symbol definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbol_arc
    """

    private: bool = False
    """The ``private`` token defines if the arc is only visible in the symbol editor. Defaults
    to ``False``.

    Available since KiCad v7"""

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token defines the coordinates of start point of the arc"""

    mid: Position = field(default_factory=lambda: Position())
    """The ``mid`` token defines the coordinates of mid point of the arc"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token defines the coordinates of end point of the arc"""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the arc outline is drawn"""

    fill: Fill = field(default_factory=lambda: Fill())
    """The ``fill`` token attributes define how the arc is filled"""

    @classmethod
    def from_sexpr(cls, exp: list) -> SyArc:
        """Convert the given S-Expresstion into a SyArc object

        Args:
            - exp (list): Part of parsed S-Expression ``(arc ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not arc

        Returns:
            - SyArc: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'arc':
            raise Exception("Expression does not have the correct type")

        object = cls()

        for item in exp:
            if isinstance(item, str):
                if item == 'private': object.private = True
                continue
            if item[0] == 'start': object.start = Position().from_sexpr(item)
            if item[0] == 'mid': object.mid = Position().from_sexpr(item)
            if item[0] == 'end': object.end = Position().from_sexpr(item)
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'fill': object.fill = Fill().from_sexpr(item)
        return object

    def to_sexpr(self, indent: int = 6, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 6.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        startA = f' {self.start.angle}' if self.start.angle is not None else ''
        midA = f' {self.mid.angle}' if self.mid.angle is not None else ''
        endA = f' {self.end.angle}' if self.end.angle is not None else ''
        private = ' private' if self.private else ''

        expression =  f'{indents}(arc{private} (start {self.start.X} {self.start.Y}{startA}) (mid {self.mid.X} {self.mid.Y}{midA}) (end {self.end.X} {self.end.Y}{endA})\n'
        expression += self.stroke.to_sexpr(indent+2)
        expression += self.fill.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class SyCircle():
    """The ``circle`` token defines a graphical circle in a symbol definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbol_circle
    """

    private: bool = False
    """The ``private`` token defines if the circle is only visible in the symbol editor. Defaults
    to ``False``.

    Available since KiCad v7"""

    center: Position = field(default_factory=lambda: Position())
    """The ``center`` token defines the coordinates of center point of the circle"""

    radius: float = 0.0
    """The ``radius`` token defines the length of the radius of the circle"""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the circle outline is drawn"""

    fill: Fill = field(default_factory=lambda: Fill())
    """The ``fill`` token attributes define how the circle is filled"""

    @classmethod
    def from_sexpr(cls, exp: list) -> SyCircle:
        """Convert the given S-Expresstion into a SyCircle object

        Args:
            - exp (list): Part of parsed S-Expression ``(circle ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not circle

        Returns:
            - SyCircle: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'circle':
            raise Exception("Expression does not have the correct type")

        object = cls()

        for item in exp:
            if isinstance(item, str):
                if item == 'private': object.private = True
                continue
            if item[0] == 'center': object.center = Position().from_sexpr(item)
            if item[0] == 'radius': object.radius = item[1]
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'fill': object.fill = Fill().from_sexpr(item)
        return object

    def to_sexpr(self, indent: int = 6, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 6.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        private = ' private' if self.private else ''

        expression =  f'{indents}(circle{private} (center {self.center.X} {self.center.Y}) (radius {self.radius})\n'
        expression += self.stroke.to_sexpr(indent+2)
        expression += self.fill.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class SyCurve():
    """The ``curve`` token defines a graphical Qubic Bezier curve.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbol_curve
    """

    points: List[Position] = field(default_factory=list)
    """The ``points`` token defines the four X/Y coordinates of each point of the curve"""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the curve outline is drawn"""

    fill: Fill = field(default_factory=lambda: Fill())
    """The ``fill`` token attributes define how curve arc is filled"""

    @classmethod
    def from_sexpr(cls, exp: list) -> SyCurve:
        """Convert the given S-Expresstion into a SyCurve object

        Args:
            - exp (list): Part of parsed S-Expression ``(curve ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not curve

        Returns:
            - SyCurve: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'curve':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'fill': object.fill = Fill().from_sexpr(item)
            if item[0] == 'pts':
                for point in item[1:]:
                    object.points.append(Position().from_sexpr(point))
        return object

    def to_sexpr(self, indent: int = 6, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 6.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(curve\n'
        expression =  f'{indents}  (pts\n'
        for point in self.points:
            expression =  f'{indents}    (xy {point.X} {point.Y})\n'
        expression =  f'{indents}  )\n'
        expression += self.stroke.to_sexpr(indent+2)
        expression += self.fill.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class SyPolyLine():
    """The ``polyline`` token defines one or more graphical lines that may or may not define a polygon.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbol_line
    """

    points: List[Position] = field(default_factory=list)
    """The ``points`` token defines the four X/Y coordinates of each point of the polyline"""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the polyline outline is drawn"""

    fill: Fill = field(default_factory=lambda: Fill())
    """The ``fill`` token attributes define how polyline arc is filled"""

    @classmethod
    def from_sexpr(cls, exp: list) -> SyPolyLine:
        """Convert the given S-Expresstion into a SyPolyLine object

        Args:
            - exp (list): Part of parsed S-Expression ``(polyline ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not polyline

        Returns:
            - SyPolyLine: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'polyline':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'fill': object.fill = Fill().from_sexpr(item)
            if item[0] == 'pts':
                for point in item[1:]:
                    object.points.append(Position().from_sexpr(point))
        return object

    def to_sexpr(self, indent: int = 6, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 6.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(polyline\n'
        expression +=  f'{indents}  (pts\n'
        for point in self.points:
            expression +=  f'{indents}    (xy {point.X} {point.Y})\n'
        expression += f'{indents}  )\n'
        expression += self.stroke.to_sexpr(indent+2)
        expression += self.fill.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class SyRect():
    """The ``rectangle`` token defines a graphical rectangle in a symbol definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbol_rectangle
    """

    private: bool = False
    """The ``private`` token defines if the rectangle is only visible in the symbol editor. Defaults
    to ``False``.
    
    Available since KiCad v7"""

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token attributes define the coordinates of the start point of the rectangle"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token attributes define the coordinates of the end point of the rectangle"""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the rectangle outline is drawn"""

    fill: Fill = field(default_factory=lambda: Fill())
    """The ``fill`` token attributes define how rectangle arc is filled"""

    @classmethod
    def from_sexpr(cls, exp: list) -> SyRect:
        """Convert the given S-Expresstion into a SyRect object

        Args:
            - exp (list): Part of parsed S-Expression ``(rectangle ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not rectangle

        Returns:
            - SyRect: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'rectangle':
            raise Exception("Expression does not have the correct type")

        object = cls()

        for item in exp:
            if isinstance(item, str):
                if item == 'private': object.private = True
                continue
            if item[0] == 'start': object.start = Position().from_sexpr(item)
            if item[0] == 'end': object.end = Position().from_sexpr(item)
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'fill': object.fill = Fill().from_sexpr(item)
        return object

    def to_sexpr(self, indent: int = 6, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 6.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        private = ' private' if self.private else ''

        expression =  f'{indents}(rectangle{private} (start {self.start.X} {self.start.Y}) (end {self.end.X} {self.end.Y})\n'
        expression += self.stroke.to_sexpr(indent+2)
        expression += self.fill.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class SyText():
    """The ``text`` token defines a graphical text in a symbol definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbol_text
    """

    text: str = ""
    """The ``text`` attribute is a quoted string that defines the text"""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y coordinates and rotation angle of the text"""

    effects: Effects = field(default_factory=lambda: Effects())
    """The ``effects`` token defines how the text is displayed"""

    @classmethod
    def from_sexpr(cls, exp: list) -> SyText:
        """Convert the given S-Expresstion into a SyText object

        Args:
            - exp (list): Part of parsed S-Expression ``(text ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not text

        Returns:
            - SyText: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'text':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.text = exp[1]
        for item in exp[2:]:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'effects': object.effects = Effects().from_sexpr(item)
        return object

    def to_sexpr(self, indent: int = 6, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 6.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''

        expression =  f'{indents}(text "{dequote(self.text)}" (at {self.position.X} {self.position.Y}{posA})\n'
        expression += f'{indents}  {self.effects.to_sexpr()}'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class SyTextBox():
    """The ``text_box`` token defines a text box inside a symbol

    Available since KiCad v7

    Documentation:
        ????
    """

    # NOTE: This is almost the same implementation as in ``kiutils.items.schitems.TextBox`` but
    #       used in a different context.

    text: str = ""
    """The ``text`` token defines the text string"""

    private: bool = False
    """The ``private`` token defines if the text box is only visible in the symbol editor. Defaults
    to ``False``."""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` token defines the X and Y coordinates and rotation angle of the text"""

    size: Position = field(default_factory=lambda: Position())
    """The ``size`` token defines the size in X and Y direction. Angle is not used."""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` token defines the look of the outline of the text box"""

    fill: Fill = field(default_factory=lambda: Fill())
    """The ``fill`` token defines how the text box should be filled"""

    effects: Effects = field(default_factory=lambda: Effects())
    """The ``effects`` token defines how the text is drawn"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    @classmethod
    def from_sexpr(cls, exp: list) -> SyTextBox:
        """Convert the given S-Expresstion into a SyTextBox object

        Args:
            - exp (list): Part of parsed S-Expression ``(text_box ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not text_box

        Returns:
            - SyTextBox: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'text_box':
            raise Exception("Expression does not have the correct type")

        object = cls()

        # Extract "private" token, if any is present
        if exp[1] == "private" and not isinstance(exp[2], list):
            object.private = True
            object.text = exp[2]
            start_at = 3
        else:
            object.text = exp[1]
            start_at = 2

        for item in exp[start_at:]:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'size': object.size = Position().from_sexpr(item)
            if item[0] == 'effects': object.effects = Effects().from_sexpr(item)
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'fill': object.fill = Fill().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''
        private = ' private' if self.private else ''

        expression =  f'{indents}(text_box{private} "{dequote(self.text)}"\n'
        expression += f'{indents}  (at {self.position.X} {self.position.Y}{posA}) (size {self.size.X} {self.size.Y})\n'
        expression += self.stroke.to_sexpr(indent+2)
        expression += self.fill.to_sexpr(indent+2)
        expression += self.effects.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression


"""Classes to manage KiCad board items

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    20.02.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List

from kiutils.items.common import Position
from kiutils.utils.strings import dequote

@dataclass
class GeneralSettings():
    """The ``general`` token define general information about the board

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/#_general_section
    """

    thickness: float = 1.6
    """The ``thickness`` token attribute defines the overall board thickness"""

    @classmethod
    def from_sexpr(cls, exp: list) -> GeneralSettings:
        """Convert the given S-Expresstion into a GeneralSettings object

        Args:
            - exp (list): Part of parsed S-Expression ``(general ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not general

        Returns:
            - GeneralSettings: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'general':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'thickness': object.thickness = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(general\n'
        expression += f'{indents}  (thickness {self.thickness})\n'
        expression += f'{indents}){endline}'
        return expression


@dataclass
class LayerToken():
    """Intermediate type used for the ``layers`` token in a board

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/#_layers_section
    """

    ordinal: int = 0
    """The layer ``ordinal`` is an integer used to associate the layer stack ordering. This is mostly
    to ensure correct mapping when the number of layers is increased in the future"""

    name: str = "F.Cu"
    """The ``name`` is the layer name defined for internal board use"""

    type: str = "signal"
    """The layer ``type`` defines the type of layer and can be defined as ``jumper``, ``mixed``, ``power``,
    ``signal``, or ``user``."""

    userName: Optional[str] = None
    """The optional ``userName`` attribute defines the custom user name"""

    @classmethod
    def from_sexpr(cls, exp: list) -> LayerToken:
        """Convert the given S-Expresstion into a LayerToken object

        Args:
            - exp (list): Part of parsed S-Expression ``(<nr> "<name>" <type>)``

        Raises:
            - Exception: When given parameter's type is not a list or the length of the list is not 3 - 4
            - Exception: When the first item of the list is not kicad_pcb

        Returns:
            - LayerToken: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) < 3 or len(exp) > 4:
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.ordinal = exp[0]
        object.name = exp[1]
        object.type = exp[2]
        if len(exp) == 4:
            object.userName = exp[3]
        return object

    def to_sexpr(self, indent=4, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        username = f' "{dequote(self.userName)}"' if self.userName is not None else ''

        return f'{indents}({self.ordinal} "{dequote(self.name)}" {self.type}{username}){endline}'


@dataclass
class StackupSubLayer():
    """The ``StackupSubLayer`` token defines a sublayer used when stacking dielectrics in a PCB"""

    thickness: float = 0.1
    """The ``thickness`` token defines the thickness of the sublayer. Defaults to 0.1"""

    material: Optional[str] = None
    """The optional ``material`` token defines a string that describes the sublayer material"""

    epsilonR: Optional[float] = None
    """The optional ``epsilonR`` token defines the dielectric constant of the sublayer material"""

    lossTangent: Optional[float] = None
    """The optional layer ``lossTangent`` token defines the dielectric loss tangent of the sublayer"""

    @classmethod
    def from_sexpr(cls, exp: list) -> StackupSubLayer:
        """This class cannot be derived from an S-Expression as the format currently used in KiCad
        board files does not match the usual convention. Assign member values manually when using
        this object.

        Raises:
            - NotImplementedError"""
        raise NotImplementedError("This class cannot be derived from an S-Expression!")

    def to_sexpr(self, indent=0, newline=False) -> str:
        """Generate the S-Expression representing this object. The representation differs from the
        normal form of an S-Expression as this uses no opening and closing parenthesis.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        mat = f' (material "{dequote(self.material)}")' if self.material is not None else ''
        er = f' (epsilon_r {self.epsilonR})' if self.epsilonR is not None else ''
        lt = f' (loss_tangent {self.lossTangent})' if self.lossTangent is not None else ''

        return f'{indents}addsublayer (thickness {self.thickness}){mat}{er}{lt}{endline}'


@dataclass
class StackupLayer():
    """The ``layer`` token defines the stack up setting of a single layer in the board stack up
    settings.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/#_stack_up_settings
    """

    name: str = ""
    """The ``name`` attribute is either one of the canonical copper or technical layer names
    or ``dielectric ID`` if it is dielectric layer"""

    # Not found in example project ...
    #number: int = 0
    """The ``number`` attribute defines the stack order of the layer"""

    type: str = ""
    """The ``type`` token defines a string that describes the layer"""

    color: Optional[str] = None
    """The optional ``color`` token defines a string that describes the layer color. This is
    only used on solder mask and silkscreen layers"""

    thickness: Optional[float] = None
    """The optional ``thickness`` token defines the thickness of the layer where appropriate"""

    material: Optional[str] = None
    """The optional ``material`` token defines a string that describes the layer material
    where appropriate"""

    epsilonR: Optional[float] = None
    """The optional ``epsilonR`` token defines the dielectric constant of the layer material"""

    lossTangent: Optional[float] = None
    """The optional layer ``lossTangent`` token defines the dielectric loss tangent of the layer"""

    subLayers: List[StackupSubLayer] = field(default_factory=list)
    """The ``sublayers`` token defines a list of zero or more sublayers that are used to create
    stacks of dielectric layers. Does not apply to copper-type layers."""

    @classmethod
    def from_sexpr(cls, exp: list) -> StackupLayer:
        """Convert the given S-Expresstion into a StackupLayer object

        Args:
            - exp (list): Part of parsed S-Expression ``(layer ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not layer

        Returns:
            - StackupLayer: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'layer':
            raise Exception("Expression does not have the correct type")

        parsingSublayer = False
        tempSublayer = StackupSubLayer()
        object = cls()
        object.name = exp[1]
        for item in exp[2:]:
            if type(item) != type([]):
                # Start parsing the layer's sublayer if the first sublayer token was found
                if item == 'addsublayer':
                    if parsingSublayer:
                        # When the ``addsublayer`` token was found a second time, the previously
                        # parsed sublayer will be appended to the list of sublayers
                        object.subLayers.append(tempSublayer)
                        tempSublayer = StackupSubLayer()
                    else:
                        # Change state of the parser to look for StackupSubLayer tokens
                        parsingSublayer = True
                continue

            # Parse the tokens of StackupSubLayer for the current sublayer
            if parsingSublayer:
                if item[0] == 'thickness': tempSublayer.thickness = item[1]
                if item[0] == 'material': tempSublayer.material = item[1]
                if item[0] == 'epsilon_r': tempSublayer.epsilonR = item[1]
                if item[0] == 'loss_tangent': tempSublayer.lossTangent = item[1]
                continue

            # Parse the normal tokens of StackupLayer token
            if item[0] == 'type': object.type = item[1]
            if item[0] == 'thickness': object.thickness = item[1]
            if item[0] == 'material': object.material = item[1]
            if item[0] == 'epsilon_r': object.epsilonR = item[1]
            if item[0] == 'loss_tangent': object.lossTangent = item[1]
            if item[0] == 'color': object.color = item[1]

        # Add the last parsed sublayer to the list, if any
        if parsingSublayer:
            object.subLayers.append(tempSublayer)

        return object

    def to_sexpr(self, indent=6, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 6.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        color = f' (color "{dequote(self.color)}")' if self.color is not None else ''
        material = f' (material "{dequote(self.material)}")' if self.material is not None else ''
        thickness = f' (thickness {self.thickness})' if self.thickness is not None else ''
        epsilon_r = f' (epsilon_r {self.epsilonR})' if self.epsilonR is not None else ''
        loss_tangent = f' (loss_tangent {self.lossTangent})' if self.lossTangent is not None else ''

        expression = f'{indents}(layer "{dequote(self.name)}" (type "{self.type}"){color}{thickness}'
        expression +=f'{material}{epsilon_r}{loss_tangent}'
        for layer in self.subLayers:
            expression += f'\n{layer.to_sexpr(indent+2)}'
        expression += f'){endline}'
        return expression

@dataclass
class Stackup():
    """The ``stackup`` token defines the board stack up settings and is defined in the setup
    section.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/#_stack_up_settings
    """

    layers: List[StackupLayer] = field(default_factory=list)
    """The ``layers``token is a list of layer settings for each layer required to manufacture
    a board including the dielectric material between the actual layers defined in the board
    editor."""

    copperFinish: Optional[str] = None
    """The optional ``copperFinish`` token is a string that defines the copper finish used to
    manufacture the board"""

    dielectricContraints: Optional[str] = None
    """The optional ``dielectricContraints`` token define if the board should meet all
    dielectric requirements. Valid values are ``yes`` and ``no``."""

    edgeConnector: Optional[str] = None
    """The optional ``edgeConnector`` token defines if the board has an edge connector
    (value: ``yes``) and if the edge connector is bevelled (value: ``bevelled``)"""

    castellatedPads: bool = False
    """The ``castellatedPads`` token defines if the board edges contain castellated pads"""

    edgePlating: bool = False
    """The ``edgePlating`` token defines if the board edges should be plated."""

    @classmethod
    def from_sexpr(cls, exp: list) -> Stackup:
        """Convert the given S-Expresstion into a Stackup object

        Args:
            - exp (list): Part of parsed S-Expression ``(stackup ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not stackup

        Returns:
            - Stackup: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'stackup':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'layer': object.layers.append(StackupLayer().from_sexpr(item))
            if item[0] == 'copper_finish': object.copperFinish = item[1]
            if item[0] == 'dielectric_constraints': object.dielectricContraints = item[1]
            if item[0] == 'edge_connector': object.edgeConnector = item[1]
            if item[0] == 'castellated_pads': object.castellatedPads = True
            if item[0] == 'edge_plating': object.edgePlating = True
        return object

    def to_sexpr(self, indent=4, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(stackup\n'
        for layer in self.layers:
            expression += layer.to_sexpr(indent+2)
        if self.copperFinish is not None:         expression += f'{indents}  (copper_finish "{dequote(self.copperFinish)}")\n'
        if self.dielectricContraints is not None: expression += f'{indents}  (dielectric_constraints {self.dielectricContraints})\n'
        if self.edgeConnector is not None:        expression += f'{indents}  (edge_connector {self.edgeConnector})\n'
        if self.castellatedPads:                  expression += f'{indents}  (castellated_pads yes)\n'
        if self.edgePlating:                      expression += f'{indents}  (edge_plating yes)\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class PlotSettings():
    """The ``pcbplotparams`` token defines the plotting and printing settings used for the last
    plot and is defined in the set up section.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/#_plot_settings
    """

    layerSelection: str = ""
    """The ``layerSelection`` token defines a hexadecimal bit set of the layers to plot"""

    plotOnAllLayersSelection: Optional[str] = None
    """The ``plotOnAllLayersSelection`` token defines a hexadecimal bit set of layers where all 
    selected layers shall be plotted.
    
    Available and required since KiCad v7"""

    disableApertMacros: bool = False
    """The ``disableApertMacros`` token defines if aperture macros are to be used in gerber plots"""

    useGerberExtensions: bool = False
    """The ``useGerberExtensions`` token  defines if the Protel layer file name extensions are to
    be used in gerber plots"""

    useGerberAttributes: bool = False
    """The ``useGerberAttributes`` token defines if the X2 extensions are used in gerber plots"""

    useGerberAdvancedAttributes: bool = False
    """The ``useGerberAdvancedAttributes`` token defines if the netlist information should be
    included in gerber plots"""

    createGerberJobFile: bool = False
    """The ``createGerberJobFile`` token defines if a job file should be created when plotting 
    gerber files"""

    # FIXME: Where is the docu of this token?
    dashedLineDashRatio: Optional[float] = None
    """The ``dashedLineDashRatio`` token's documentation is still missing ..
    
    Available and required since KiCad v7"""

    # FIXME: Where is the docu of this token?
    dashedLineGapRatio: Optional[float] = None
    """The ``dashedLineGapRatio`` token's documentation is still missing ..
    
    Available and required since KiCad v7"""

    svgUseInch: Optional[bool] = None
    """The ``svgUseInch`` token defines if inch units should be use when plotting SVG files.
    
    Required until KiCad v6, removed since KiCad v7"""

    svgPrecision: float = 0.0
    """The ``svgPrecision`` token defines the units precision used when plotting SVG files"""

    excludeEdgeLayer: Optional[bool] = None
    """The ``excludeEdgeLayer`` token defines if the board edge layer is plotted on all layers.
    
    Required until KiCad v6, removed since KiCad v7"""

    plotFameRef: bool = False
    """The ``plotFameRef`` token defines if the border and title block should be plotted"""

    viasOnMask: bool = False
    """The ``viasOnMask`` token defines if the vias are to be tented"""

    mode: int = 1
    """The ``mode`` token defines the plot mode. An attribute of 1 plots in the normal
    mode and an attribute of 2 plots in the outline (sketch) mode."""

    useAuxOrigin: bool = False
    """The ``useAuxOrigin`` token determines if all coordinates are offset by the defined user origin"""

    hpglPenNumber: int = 0
    """The ``hpglPenNumber`` token defines the integer pen number used for HPGL plots"""

    hpglPenSpeed: int = 0
    """The ``hpglPenSpeed`` token defines the integer pen speed used for HPGL plots"""

    hpglPenDiameter: float = 0.0
    """The ``hpglPenDiameter`` token defines the floating point pen size for HPGL plots"""

    dxfPolygonMode: bool = False
    """The ``dxfPolygonMode`` token defines if the polygon mode should be used for DXF plots"""

    dxfImperialUnits: bool = False
    """The ``dxfImperialUnits`` token defines if imperial units should be used for DXF plots"""

    dxfUsePcbnewFont: bool = False
    """The ``dxfUsePcbnewFont`` token defines if the Pcbnew font (vector font) or the default
    font should be used for DXF plots"""

    psNegative: bool = False
    """The ``psNegative`` token defines if the output should be the negative for PostScript plots"""

    psA4Output: bool = False
    """The ``psA4Output`` token defines if the A4 page size should be used for PostScript plots"""

    plotReference: bool = False
    """The ``plotReference`` token defines if hidden reference field text should be plotted"""

    plotValue: bool = False
    """The ``plotValue`` token defines if hidden value field text should be plotted"""

    plotInvisibleText: bool = False
    """The ``plotInvisibleText`` token defines if hidden text other than the reference and
    value fields should be plotted"""

    sketchPadsOnFab: bool = False
    """The ``sketchPadsOnFab`` token defines if pads should be plotted in the outline (sketch) mode"""

    subtractMaskFromSilk: bool = False
    """The ``subtractMaskFromSilk`` token defines if the solder mask layers should be subtracted from
    the silk screen layers for gerber plots"""

    outputFormat: int = 0
    """The ``outputFormat`` token defines the last plot type. The following values are defined:
    - 0: gerber
    - 1: PostScript
    - 2: SVG
    - 3: DXF
    - 4: HPGL
    - 5: PDF"""

    mirror: bool = False
    """The ``mirror`` token defines if the plot should be mirrored"""

    drillShape: int = 0
    """The ``drillShape`` token defines the type of drill marks used for drill files"""

    scaleSelection: int = 1
    """The ``scaleSelection`` is not documented yet (as of 20.02.2022)"""

    outputDirectory: str = ""
    """The ``drillShape`` token defines the path relative to the current project path
    where the plot files will be saved"""

    @classmethod
    def from_sexpr(cls, exp: list) -> PlotSettings:
        """Convert the given S-Expresstion into a PlotSettings object

        Args:
            - exp (list): Part of parsed S-Expression ``(pcbplotparams ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not pcbplotparams

        Returns:
            - PlotSettings: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'pcbplotparams':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'layerselection': object.layerSelection = item[1]
            if item[0] == 'plot_on_all_layers_selection': object.plotOnAllLayersSelection = item[1]
            if item[0] == 'disableapertmacros': object.disableApertMacros = True if item[1] == 'true' else False
            if item[0] == 'usegerberextensions' : object.useGerberExtensions = True if item[1] == 'true' else False
            if item[0] == 'usegerberattributes' : object.useGerberAttributes = True if item[1] == 'true' else False
            if item[0] == 'usegerberadvancedattributes' : object.useGerberAdvancedAttributes = True if item[1] == 'true' else False
            if item[0] == 'creategerberjobfile' : object.createGerberJobFile = True if item[1] == 'true' else False
            if item[0] == 'dashed_line_dash_ratio': object.dashedLineDashRatio = item[1]
            if item[0] == 'dashed_line_gap_ratio': object.dashedLineGapRatio = item[1]
            if item[0] == 'svguseinch' : object.svgUseInch = True if item[1] == 'true' else False
            if item[0] == 'svgprecision' : object.svgPrecision = item[1]
            if item[0] == 'excludeedgelayer' : object.excludeEdgeLayer = True if item[1] == 'true' else False
            if item[0] == 'plotframeref' : object.plotFameRef = True if item[1] == 'true' else False
            if item[0] == 'viasonmask' : object.viasOnMask = True if item[1] == 'true' else False
            if item[0] == 'mode' : object.mode = item[1]
            if item[0] == 'useauxorigin' : object.useAuxOrigin = True if item[1] == 'true' else False
            if item[0] == 'hpglpennumber' : object.hpglPenNumber = item[1]
            if item[0] == 'hpglpenspeed' : object.hpglPenSpeed = item[1]
            if item[0] == 'hpglpendiameter' : object.hpglPenDiameter = item[1]
            if item[0] == 'dxfpolygonmode' : object.dxfPolygonMode = True if item[1] == 'true' else False
            if item[0] == 'dxfimperialunits' : object.dxfImperialUnits = True if item[1] == 'true' else False
            if item[0] == 'dxfusepcbnewfont' : object.dxfUsePcbnewFont = True if item[1] == 'true' else False
            if item[0] == 'psnegative' : object.psNegative = True if item[1] == 'true' else False
            if item[0] == 'psa4output' : object.psA4Output = True if item[1] == 'true' else False
            if item[0] == 'plotreference' : object.plotReference = True if item[1] == 'true' else False
            if item[0] == 'plotvalue' : object.plotValue = True if item[1] == 'true' else False
            if item[0] == 'plotinvisibletext' : object.plotInvisibleText = True if item[1] == 'true' else False
            if item[0] == 'sketchpadsonfab' : object.sketchPadsOnFab = True if item[1] == 'true' else False
            if item[0] == 'subtractmaskfromsilk' : object.subtractMaskFromSilk = True if item[1] == 'true' else False
            if item[0] == 'outputformat' : object.outputFormat = item[1]
            if item[0] == 'mirror' : object.mirror = True if item[1] == 'true' else False
            if item[0] == 'drillshape' : object.drillShape = item[1]
            if item[0] == 'scaleselection' : object.scaleSelection = item[1]
            if item[0] == 'outputdirectory' : object.outputDirectory = item[1]
        return object

    def to_sexpr(self, indent=4, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(pcbplotparams\n'
        expression += f'{indents}  (layerselection {self.layerSelection})\n'
        if self.plotOnAllLayersSelection is not None:
            expression += f'{indents}  (plot_on_all_layers_selection {self.plotOnAllLayersSelection})\n'
        expression += f'{indents}  (disableapertmacros {str(self.disableApertMacros).lower()})\n'
        expression += f'{indents}  (usegerberextensions {str(self.useGerberExtensions).lower()})\n'
        expression += f'{indents}  (usegerberattributes {str(self.useGerberAttributes).lower()})\n'
        expression += f'{indents}  (usegerberadvancedattributes {str(self.useGerberAdvancedAttributes).lower()})\n'
        expression += f'{indents}  (creategerberjobfile {str(self.createGerberJobFile).lower()})\n'
        if self.dashedLineDashRatio is not None:
            expression += f'{indents}  (dashed_line_dash_ratio {float(self.dashedLineDashRatio):.6f})\n'
        if self.dashedLineGapRatio is not None:
            expression += f'{indents}  (dashed_line_gap_ratio {float(self.dashedLineGapRatio):.6f})\n'
        if self.svgUseInch is not None:
            expression += f'{indents}  (svguseinch {str(self.svgUseInch).lower()})\n'
        expression += f'{indents}  (svgprecision {self.svgPrecision})\n'
        if self.excludeEdgeLayer is not None:
            expression += f'{indents}  (excludeedgelayer {str(self.excludeEdgeLayer).lower()})\n'
        expression += f'{indents}  (plotframeref {str(self.plotFameRef).lower()})\n'
        expression += f'{indents}  (viasonmask {str(self.viasOnMask).lower()})\n'
        expression += f'{indents}  (mode {self.mode})\n'
        expression += f'{indents}  (useauxorigin false)\n'
        expression += f'{indents}  (hpglpennumber {self.hpglPenNumber})\n'
        expression += f'{indents}  (hpglpenspeed {self.hpglPenSpeed})\n'
        expression += f'{indents}  (hpglpendiameter {float(self.hpglPenDiameter):.6f})\n'
        expression += f'{indents}  (dxfpolygonmode {str(self.dxfPolygonMode).lower()})\n'
        expression += f'{indents}  (dxfimperialunits {str(self.dxfImperialUnits).lower()})\n'
        expression += f'{indents}  (dxfusepcbnewfont {str(self.dxfUsePcbnewFont).lower()})\n'
        expression += f'{indents}  (psnegative {str(self.psNegative).lower()})\n'
        expression += f'{indents}  (psa4output {str(self.psA4Output).lower()})\n'
        expression += f'{indents}  (plotreference {str(self.plotReference).lower()})\n'
        expression += f'{indents}  (plotvalue {str(self.plotValue).lower()})\n'
        expression += f'{indents}  (plotinvisibletext {str(self.plotInvisibleText).lower()})\n'
        expression += f'{indents}  (sketchpadsonfab {str(self.sketchPadsOnFab).lower()})\n'
        expression += f'{indents}  (subtractmaskfromsilk {str(self.subtractMaskFromSilk).lower()})\n'
        expression += f'{indents}  (outputformat {self.outputFormat})\n'
        expression += f'{indents}  (mirror {str(self.mirror).lower()})\n'
        expression += f'{indents}  (drillshape {self.drillShape})\n'
        expression += f'{indents}  (scaleselection {self.scaleSelection})\n'
        expression += f'{indents}  (outputdirectory "{dequote(self.outputDirectory)}")\n'
        expression += f'{indents}){endline}'
        return expression


@dataclass
class SetupData():
    """The setup token is used to store the current settings such as default item sizes and
    other options used by the board

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/#_setup_section
    """

    stackup: Optional[Stackup] = None
    """The optional ``stackup`` define the parameters required to manufacture the board"""

    packToMaskClearance: float = 0.0
    """The ``packToMaskClearance`` token defines the clearance between footprint pads and
    the solder mask"""

    solderMaskMinWidth: Optional[float] = None
    """The optional ``solderMaskMinWidth`` defines the minimum solder mask width. If not
    defined, the minimum width is zero."""

    padToPasteClearance: Optional[float] = None
    """The optional ``padToPasteClearance`` defines the clearance between footprint pads
    and the solder paste layer. If not defined, the clearance is zero"""

    padToPasteClearanceRatio: Optional[float] = None
    """The optional ``padToPasteClearanceRatio`` is the percentage (from 0 to 100) of the
    footprint pad to make the solder paste. If not defined, the ratio is 100% (the same
    size as the pad)."""

    auxAxisOrigin: Optional[Position] = None
    """The optional ``auxAxisOrigin`` defines the auxiliary origin if it is set to anything
    other than (0,0)."""

    gridOrigin: Optional[Position] = None
    """The optional ``gridOrigin`` defines the grid original if it is set to anything other
    than (0,0)."""

    plotSettings: Optional[PlotSettings] = None
    """The optional ``plotSettings`` define how the board was last plotted."""

    @classmethod
    def from_sexpr(cls, exp: list) -> SetupData:
        """Convert the given S-Expresstion into a SetupData object

        Args:
            - exp (list): Part of parsed S-Expression ``(setup ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not setup

        Returns:
            - SetupData: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'setup':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'stackup': object.stackup = Stackup().from_sexpr(item)
            if item[0] == 'pcbplotparams': object.plotSettings = PlotSettings().from_sexpr(item)
            if item[0] == 'pad_to_mask_clearance': object.packToMaskClearance = item[1]
            if item[0] == 'solder_mask_min_width': object.solderMaskMinWidth = item[1]
            if item[0] == 'pad_to_paste_clearance': object.padToPasteClearance = item[1]
            if item[0] == 'pad_to_paste_clearance_ratio': object.padToPasteClearanceRatio = item[1]
            if item[0] == 'aux_axis_origin': object.auxAxisOrigin = Position().from_sexpr(item)
            if item[0] == 'grid_origin': object.gridOrigin = Position().from_sexpr(item)
            if item[0] == 'pcbplotparams': object.plotSettings = PlotSettings().from_sexpr(item)
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(setup\n'
        if self.stackup is not None:                  expression += self.stackup.to_sexpr(indent+2)
        expression += f'{indents}  (pad_to_mask_clearance {self.packToMaskClearance})\n'
        if self.solderMaskMinWidth is not None:       expression += f'{indents}  (solder_mask_min_width {self.solderMaskMinWidth})\n'
        if self.padToPasteClearance is not None:      expression += f'{indents}  (pad_to_paste_clearance {self.padToPasteClearance})\n'
        if self.padToPasteClearanceRatio is not None: expression += f'{indents}  (pad_to_paste_clearance_ratio {self.padToPasteClearanceRatio})\n'
        if self.auxAxisOrigin is not None:            expression += f'{indents}  (aux_axis_origin {self.auxAxisOrigin.X} {self.auxAxisOrigin.Y})\n'
        if self.gridOrigin is not None:               expression += f'{indents}  (grid_origin {self.gridOrigin.X} {self.gridOrigin.Y})\n'
        if self.plotSettings is not None:             expression += self.plotSettings.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression


@dataclass
class Segment():
    """The ``segment`` token defines a track segment in a KiCad board

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/#_track_segment
    """

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token defines the coordinates of the beginning of the line"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token defines the coordinates of the end of the line"""

    width: float = 0.1
    """The ``width`` token defines the line width"""

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the track segment resides on"""

    locked: bool = False
    """The ``locked`` token defines if the line cannot be edited"""

    net: int = 0
    """The ``net`` token defines by the net ordinal number which net in the net
    section that the segment is part of"""

    tstamp: str = ""
    """The ``tstamp`` token defines the unique identifier of the line object"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Segment:
        """Convert the given S-Expresstion into a Segment object

        Args:
            - exp (list): Part of parsed S-Expression ``(segment ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not segment

        Returns:
            - Segment: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'segment':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                continue
            if item[0] == 'start': object.start = Position().from_sexpr(item)
            if item[0] == 'end': object.end = Position().from_sexpr(item)
            if item[0] == 'width': object.width = item[1]
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'net': object.net = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        locked = ' locked' if self.locked else ''

        return f'{indents}(segment{locked} (start {self.start.X} {self.start.Y}) (end {self.end.X} {self.end.Y}) (width {self.width}) (layer "{dequote(self.layer)}") (net {self.net}) (tstamp {self.tstamp})){endline}'

@dataclass
class Via():
    """The ``via`` token defines a track via in a KiCad board

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/#_track_via
    """

    type: Optional[str] = None
    """The optional ``type`` attribute specifies the via type. Valid via types are ``blind`` and
    ``micro``. If no type is defined, the via is a through hole type"""

    locked: bool = False
    """The ``locked`` token defines if the line cannot be edited"""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` token define the coordinates of the center of the via"""

    size: float = 0.0
    """The ``size`` token define the diameter of the via annular ring"""

    drill: float = 0.0
    """The ``drill`` token define the drill diameter of the via"""

    layers: List[str] = field(default_factory=list)
    """The ``layers`` token define the canonical layer set the via connects as a list
    of strings"""

    removeUnusedLayers: bool = False
    """The ``removeUnusedLayers`` token is undocumented (as of 20.02.2022)"""

    keepEndLayers: bool = False
    """The ``keepEndLayers`` token is undocumented (as of 20.02.2022)"""

    free: bool = False
    """The ``free`` token indicates that the via is free to be moved outside it's assigned net"""

    net: int = 0
    """The ``net`` token defines by net ordinal number which net in the net section that
    the via is part of"""

    tstamp: Optional[str] = None
    """The ``tstamp`` token defines the unique identifier of the via"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Via:
        """Convert the given S-Expresstion into a Via object

        Args:
            - exp (list): Part of parsed S-Expression ``(via ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not via

        Returns:
            - Via: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'via':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                if item == 'micro' or item == 'blind': object.type = item
                continue
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'size': object.size = item[1]
            if item[0] == 'drill': object.drill = item[1]
            if item[0] == 'layers':
                for layer in item[1:]:
                    object.layers.append(layer)
            if item[0] == 'remove_unused_layers': object.removeUnusedLayers = True
            if item[0] == 'keep_end_layers': object.keepEndLayers = True
            if item[0] == 'free': object.free = True
            if item[0] == 'net': object.net = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        type = f' {self.type}' if self.type is not None else ''
        locked = f' locked' if self.locked else ''

        layers = ''
        for layer in self.layers:
            layers += f' "{dequote(layer)}"'
        rum = f' (remove_unused_layers)' if self.removeUnusedLayers else ''
        kel = f' (keep_end_layers)' if self.keepEndLayers else ''
        free = f' (free)' if self.free else ''
        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''

        return f'{indents}(via{type}{locked} (at {self.position.X} {self.position.Y}) (size {self.size}) (drill {self.drill}) (layers{layers}){rum}{kel}{free} (net {self.net}){tstamp}){endline}'

@dataclass
class Arc():
    """The ``arc`` token defines a track arc, which will be generated when using the length-matching
    feature on differential pairs.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-pcb/#_track_arc
    """

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token defines the coordinates of the beginning of the arc"""

    mid: Position = field(default_factory=lambda: Position())
    """The ``mid`` token defines the coordinates of the mid point of the radius of the arc"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token defines the coordinates of the end of the arc"""

    width: float = 0.2
    """The ``width`` token defines the line width of the arc. Defaults to 0,2."""

    layer: str = "F.Cu"
    """The ``layer`` token defiens the canonical layer the track arc resides on. Defaults to `F.Cu`."""

    locked: bool = False
    """The ``locked`` token defines if the arc cannot be edited. Defaults to False."""

    net: int = 0
    """The ``net`` token defines the net ordinal number which net in the net section that arc is part
    of. Defaults to 0."""

    tstamp: Optional[str] = None
    """The optional ``tstamp`` token defines the unique identifier of the arc"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Arc:
        """Convert the given S-Expresstion into a Arc object

        Args:
            - exp (list): Part of parsed S-Expression ``(arc ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not ``arc``

        Returns:
            - Arc: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'arc':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                continue
            if item[0] == 'start': object.start = Position().from_sexpr(item)
            elif item[0] == 'mid': object.mid = Position().from_sexpr(item)
            elif item[0] == 'end': object.end = Position().from_sexpr(item)
            elif item[0] == 'width': object.width = item[1]
            elif item[0] == 'layer': object.layer = item[1]
            elif item[0] == 'net': object.net = item[1]
            elif item[0] == 'tstamp': object.tstamp = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        locked = f' locked' if self.locked else ''
        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''

        expression = f'{indents}(arc{locked} (start {self.start.X} {self.start.Y}) '
        expression += f'(mid {self.mid.X} {self.mid.Y}) (end {self.end.X} {self.end.Y}) '
        expression += f'(width {self.width}) (layer "{dequote(self.layer)}") '
        expression += f'(net {self.net}){tstamp}){endline}'
        return expression


@dataclass
class Target():
    """The ``target`` token defines a target marker on the PCB

    Documentation:
        Not found in KiCad docu - 15.06.2022
    """

    type: str = "plus"
    """The ``type`` token specifies the shape of the marker. Valid types are ``plus`` and ``x``."""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` token specifies the position of the target marker"""

    size: float = 0
    """The ``size`` token sets the marker's size"""

    width: float = 0.1
    """The ``width`` token sets the marker's line width"""

    layer: str = "F.Cu"
    """The ``layer`` token sets the canonical layer where the target marker resides"""

    tstamp: Optional[str] = None
    """The ``tstamp`` token defines the unique identifier of the target"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Target:
        """Convert the given S-Expresstion into a Target object

        Args:
            - exp (list): Part of parsed S-Expression ``(target ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not target

        Returns:
            - Target: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'target':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.type = exp[1]
        for item in exp[2:]:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'size': object.size = item[1]
            if item[0] == 'width': object.width = item[1]
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        return f'{indents}(target {self.type} (at {self.position.X} {self.position.Y}) (size {self.size}) (width {self.width}) (layer "{self.layer}") (tstamp {self.tstamp})){endline}'



"""Unittests of library symbol related classes

Authors:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0
"""

import unittest
from os import path

from tests.testfunctions import to_file_and_compare, prepare_test, cleanup_after_test, TEST_BASE
from kiutils.symbol import SymbolLib, Symbol
from kiutils.misc.config import KIUTILS_CREATE_NEW_VERSION_STR
from kiutils.utils.sexpr import parse_sexp

SYMBOL_BASE = path.join(TEST_BASE, 'symbol')

class Tests_Symbol(unittest.TestCase):
    """Test cases for Symbols"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_allSymbolPinVariations(self):
        """Tests the parsing of all pin types of a symbol in a symbol library"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'test_allSymbolPinVariations')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_allSymbolAlternatePins(self):
        """Tests the parsing of all alternate pin definitions of a symbol in a symbol library"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'test_allSymbolAlternatePins')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_symbolParameters(self):
        """Tests the parsing of a symbol's parameters in a symbol library"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'test_symbolParameters')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_symbolDemorganUnits(self):
        """Tests the parsing of a symbol's de-morgan unit represenations in a symbol library"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'test_symbolDemorganUnits')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_symbolDemorganSyItems(self):
        """Tests the parsing of a symbol' in a symbol library that has all SyItems in different
        de-morgan variations in it"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'test_symbolDemorganSyItems')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_bigSymbolLibrary(self):
        """Tests the parsing of a big symbol library with many symbols of different kinds in it"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'test_bigSymbolLibrary')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_createNewSymbolInEmptyLibrary(self):
        """Tests the ``create_new()`` function to create an empty symbol that is added to a
        symbol library"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'test_createNewSymbolInEmptyLibrary')

        # Create an empty symbol library
        symbolLib = SymbolLib(
            version = KIUTILS_CREATE_NEW_VERSION_STR,
            generator = 'kiutils'
        )

        # Add a symbol to it via create_new()
        symbol = Symbol().create_new(
            id = 'testsymbol',
            reference = 'U',
            value = 'testvalue'
        )
        symbolLib.symbols.append(symbol)

        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_renameParentIdUsingIdToken(self):
        """Rename symbol inside library using the id token setter and verify all units are renamed
        correctly as well as the ``Value`` property stayed the same."""
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'test_renameParentIdUsingIdToken')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        symbolLib.symbols[0].libId = 'ExampleLibrary:AD2023'    # Setting library nickname
        symbolLib.symbols[1].libId = 'AD2023'                   # Unsetting library nickname
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_createNewTopLevelSymbolFromChild(self):
        """Take a child symbol, rename its library id and make a new top-level symbol out of it.
        Tests if resetting both ``unitId`` and ``styleId`` works."""
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'test_createNewTopLevelSymbolFromChild')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)

        # Copy the symbol
        childSymbol = Symbol.from_sexpr(parse_sexp(symbolLib.symbols[0].units[0].to_sexpr()))

        # Rename it and save it as a new top-level symbol to the library
        childSymbol.libId = "SomeNewName:AD2023"
        symbolLib.symbols.append(childSymbol)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_mergeLibraries(self):
        """Merge the symbols of two symbol libraries together"""
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'test_mergeLibraries')
        symbolLib1 = SymbolLib().from_file(self.testData.pathToTestFile)
        symbolLib2 = SymbolLib().from_file(path.join(SYMBOL_BASE, 'test_symbolDemorganSyItems'))
        for symbol in symbolLib2.symbols:
            symbolLib1.symbols.insert(0, symbol)
        self.assertTrue(to_file_and_compare(symbolLib1, self.testData))

    def test_symbolIdParser(self):
        """Tests edge cases of parsing the symbol ID token and checks if the ID was split into
        its parts correctly

        Related issues:
            - [Pull request 73](https://github.com/mvnmgrx/kiutils/pull/73)
        """
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'test_symbolIdParser')
        self.testData.compareToTestFile = True
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

class Tests_Symbol_Since_V7(unittest.TestCase):
    """Test cases for Symbols since KiCad 7"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_textBoxAllVariants(self):
        """Tests all variants of the ``text_box`` token for text boxes in symbols"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'since_v7', 'test_textBoxAllVariants')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_rectangleAllVariants(self):
        """Tests all variants of the ``rectangle`` token for rectangles in symbols"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'since_v7', 'test_rectangleAllVariants')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_circleAllVariants(self):
        """Tests all variants of the ``circle`` token for circles in symbols"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'since_v7', 'test_circleAllVariants')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))

    def test_arcAllVariants(self):
        """Tests all variants of the ``arc`` token for arcs in symbols"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SYMBOL_BASE, 'since_v7', 'test_arcAllVariants')
        symbolLib = SymbolLib().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(symbolLib, self.testData))


"""Unittests for misc stuff

Authors:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0
"""

import unittest
from os import path
from kiutils.schematic import Schematic

from tests.testfunctions import to_file_and_compare, prepare_test, TEST_BASE

MISC_BASE = path.join(TEST_BASE, 'misc')

class Tests_Misc(unittest.TestCase):
    """Misc test cases"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_quotesAndBackslashInSexpr(self):
        """Tests the correct parsing of all variants of quoted strings as well as of backslashes
        in said strings.

        Related issues:
            - [Pull request 73](https://github.com/mvnmgrx/kiutils/pull/73)
        """
        self.testData.pathToTestFile = path.join(MISC_BASE, 'test_quotesAndBackslashInSexpr')
        self.testData.compareToTestFile = True
        libtable = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(libtable, self.testData))
